"""
Update existing collection.

TODO: Refactor, this script was written in one setting and needs to be revised significantly.
"""

# import argparse
# import logging

# from ripple1d.utils.s3_utils import read_json_from_s3

# from .fim_collection import FIMCollection, FIMCollectionRasItem

# if __name__ == "__main__":
#     # STAC API URL
#     API_URL = "https://stac2.dewberryanalytics.com"

#     parser = argparse.ArgumentParser(description="Process some integers.")
#     parser.add_argument("--collection_id", type=str, required=True, help="Collection ID")

#     args = parser.parse_args()
#     collection_id = args.collection_id

#     fc = FIMCollection(API_URL, collection_id)
#     for item in fc.collection.get_all_items():
#         logging.info(f"item_id: {item.id}")

#         fci = FIMCollectionRasItem(API_URL, collection_id, item.id)

#         fci.sanitize_ras_stac_props()

#         fci.map_topo_assets()

#         conflated = fci.add_ripple1d_params()

#         fci.add_s3_key_to_assets()

#         fci.ensure_asset_roles_unique()

#         if conflated:
#             params_json = fci.item.assets["ripple1d_parameters.json"].extra_fields["s3_key"]
#             data = read_json_from_s3("fim", params_json)

#             fci.item.properties["FIM:Branch Metadata"] = data
#         else:
#             logging.error(f"Failed to add ripple1d params for {fci.item.id}")

#         fci.post_item_updates()



"""Utils for working with raster data."""

import json
import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Tuple

import numpy as np
import pystac
import rasterio
import rasterio.warp
import shapely
import shapely.ops
from mypy_boto3_s3.service_resource import Object
from pyproj import CRS
from rasterio import mask
from rasterio.session import AWSSession
from rasterio.warp import Resampling, calculate_default_transform, reproject
from shapely import Polygon

from ripple1d.consts import METERS_PER_FOOT
from ripple1d.errors import UnknownVerticalUnits

from .s3_utils import *

logging.getLogger("boto3").setLevel(logging.WARNING)
logging.getLogger("botocore").setLevel(logging.WARNING)


def get_raster_bounds(raster_file: str) -> Tuple[float, float, float, float]:
    """
    Retrieve the geographic bounds of a raster file and returns them in the WGS 84 (EPSG:4326) coordinate reference system.

    Parameters
    ----------
        raster_file (str): The raster file.

    Returns
    -------
        Tuple[float, float, float, float]: The geographic bounds of the raster file in the WGS 84 (EPSG:4326) coordinate reference system. The bounds are returned as a tuple of four floats: (west, south, east, north).
    """
    with rasterio.open(raster_file) as src:
        bounds = src.bounds
        crs = src.crs
        bounds_4326 = rasterio.warp.transform_bounds(crs, "EPSG:4326", *bounds)
        return bounds_4326


def get_raster_metadata(raster_file: str) -> dict:
    """
    Retrieve the metadata of a raster file.

    Parameters
    ----------
        raster_file (str): The raster file.

    Returns
    -------
        dict: The metadata of the raster file. The metadata is returned as a dictionary
        where the keys are the names of the metadata items and the values are the values of the metadata items.
    """
    with rasterio.open(raster_file) as src:
        return src.tags(1)


def bbox_to_polygon(bbox) -> shapely.Polygon:
    """
    Convert a bounding box to a Shapely Polygon.

    Parameters
    ----------
        bbox: The bounding box to convert. It should be a sequence of four numbers: (min_x, min_y, max_x, max_y).

    Returns
    -------
        shapely.Polygon: The Shapely Polygon representing the bounding box. The Polygon is a rectangle with the lower
          left corner at (min_x, min_y) and the upper right corner at (max_x, max_y).
    """
    min_x, min_y, max_x, max_y = bbox
    return shapely.Polygon(
        [
            [min_x, min_y],
            [min_x, max_y],
            [max_x, max_y],
            [max_x, min_y],
        ]
    )


def get_unit_name(crs: CRS):
    """Get units from a crs object."""
    unit_name = crs.axis_info[0].unit_name
    english = ["ft", "feet", "foot"]
    metric = ["m", "meter", "meters"]
    for name in english:
        if name in unit_name.lower():
            return "Feet"
    for name in metric:
        if name in unit_name.lower():
            return "Meters"
    raise ValueError(f"unrecognized units ")


def convert_units(dst_crs: CRS, resolution: float, resolution_units: str) -> float:
    """Convert resolution to match the units of the destination crs."""
    dest_units = get_unit_name(dst_crs)
    if dest_units != resolution_units:
        if resolution_units == "Feet" and dest_units == "Meters":
            resolution = resolution * METERS_PER_FOOT
        elif resolution_units == "Meters" and dest_units == "Feet":
            resolution = resolution / METERS_PER_FOOT
    return resolution


def reproject_raster(
    src_path: str,
    dest_path: str,
    dst_crs: CRS,
    resolution: float = None,
    resolution_units: str = None,
    COMPRESS="DEFLATE",
    PREDICTOR="3",
    num_threads=4,
    tiled=False,
    blocksize=512,
):
    """Reproject/resample raster."""
    with rasterio.open(src_path) as src:
        if not resolution and not resolution_units:
            resolution = src.res[0]
            transform, width, height = calculate_default_transform(src.crs, dst_crs, src.width, src.height, *src.bounds)
        else:
            resolution = convert_units(dst_crs, resolution, resolution_units)
            transform, width, height = calculate_default_transform(
                src.crs, dst_crs, src.width, src.height, *src.bounds, resolution=resolution
            )
        kwargs = src.meta.copy()
        kwargs.update({"crs": dst_crs, "transform": transform, "width": width, "height": height})

        logging.info(resolution)
        kwargs.update(
            {
                "blockysize": blocksize,
                "blockxsize": blocksize,
                "tiled": tiled,
                "COMPRESS": COMPRESS,
                "PREDICTOR": PREDICTOR,
                "num_threads": num_threads,
            }
        )
        with rasterio.open(dest_path, "w", **kwargs) as dst:
            for i in range(1, src.count + 1):
                reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=transform,
                    dst_crs=dst_crs,
                    dst_resolution=resolution,
                    resampling=Resampling.nearest,
                    kwarg=kwargs,
                )


def clip_raster(src_path: str, dst_path: str, mask_polygon: Polygon, vertical_units: str):
    """Clip a raster file to a polygon and save the result to a new file."""
    if os.path.exists(dst_path):
        raise FileExistsError(dst_path)
    if not isinstance(mask_polygon, Polygon):
        raise TypeError(mask_polygon)
    os.makedirs(os.path.dirname(dst_path), exist_ok=True)

    logging.info(f"Reading: {src_path}")
    with rasterio.open(src_path) as src:
        out_meta = src.meta
        out_image, out_transform = mask.mask(src, [mask_polygon], all_touched=True, crop=True)

    out_meta.update(
        {
            "driver": "GTiff",
            "height": out_image.shape[1],
            "width": out_image.shape[2],
            "transform": out_transform,
            "compress": "LZW",
            "predictor": 3,
            "tiled": True,
        }
    )

    logging.info(f"Writing as masked: {dst_path}")
    with rasterio.open(dst_path, "w", **out_meta) as dest:
        if vertical_units == "Meters":
            out_image_ft = np.where(out_image == out_meta["nodata"], out_meta["nodata"], out_image / METERS_PER_FOOT)
            dest.write(out_image_ft)
        elif vertical_units == "Feet":
            dest.write(out_image)
        else:
            raise UnknownVerticalUnits(f"Expected Feet or Meters recieved {vertical_units}")


def get_terrain_exe_path(ras_ver: str) -> str:
    """Return Windows path to RasProcess.exe exposing CreateTerrain subroutine, compatible with provided RAS version."""
    # 5.0.7 version of RasProcess.exe does not expose CreateTerrain subroutine.
    # Testing shows that RAS 5.0.7 accepts Terrain created by 6.1 version of RasProcess.exe, so use that for 5.0.7.
    d = {
        "507": r"C:\Program Files (x86)\HEC\HEC-RAS\6.1\RasProcess.exe",
        "5.07": r"C:\Program Files (x86)\HEC\HEC-RAS\6.1\RasProcess.exe",
        "600": r"C:\Program Files (x86)\HEC\HEC-RAS\6.0\RasProcess.exe",
        "6.00": r"C:\Program Files (x86)\HEC\HEC-RAS\6.0\RasProcess.exe",
        "610": r"C:\Program Files (x86)\HEC\HEC-RAS\6.1\RasProcess.exe",
        "6.10": r"C:\Program Files (x86)\HEC\HEC-RAS\6.1\RasProcess.exe",
        "631": r"C:\Program Files (x86)\HEC\HEC-RAS\6.3.1\RasProcess.exe",
        "6.3.1": r"C:\Program Files (x86)\HEC\HEC-RAS\6.3.1\RasProcess.exe",
    }
    try:
        return d[ras_ver]
    except KeyError as e:
        raise ValueError(f"Unsupported ras_ver: {ras_ver}. choices: {sorted(d)}") from e



"""Utils for working with STAC items/catalogs."""

import json
import logging
import os
from datetime import datetime
from pathlib import Path
from typing import List
from urllib.parse import quote

import boto3
import pystac
import pystac_client
import requests
from ripple1d.utils.s3_utils import s3_get_output_s3path


def key_to_uri(key: str, bucket: str) -> str:
    """Convert a key to a uri."""
    return f"https://{bucket}.s3.amazonaws.com/{quote(key)}"


def uri_to_key(href: str, bucket: str) -> str:
    """Convert a uri to a key."""
    return href.replace(f"https://{bucket}.s3.amazonaws.com/", "")


def collection_exists(endpoint: str, collection_id: str):
    """Check if a collection exists in a STAC API."""
    return requests.get(f"{endpoint}/collections/{collection_id}")


def create_collection(
    models: List[pystac.Item], id: str, description: str = None, title: str = None
) -> pystac.Collection:
    """Create a STAC collection from a list of STAC items."""
    extent = pystac.Extent.from_items(models)
    return pystac.Collection(
        id=id,
        description=description,
        title=title,
        extent=extent,
    )


def upsert_collection(endpoint: str, collection: pystac.Collection, headers: dict):
    """Upsert a collection to a STAC API."""
    collections_url = f"{endpoint}/collections"
    response = requests.post(collections_url, json=collection.to_dict(), headers=headers)
    if response.status_code == 409:
        logging.warning("collection already exists, updating...")
        collections_update_url = f"{collections_url}/{collection.id}"
        response = requests.put(collections_update_url, json=collection.to_dict(), headers=headers)
        if response.status_code != 200:
            raise RuntimeError(f"Error putting collection {(response.status_code)}")
    elif response.status_code not in (201, 200):
        raise RuntimeError(f"Error posting collection {(response.status_code)}")


def upsert_item(endpoint: str, collection_id: str, item: pystac.Item, headers: dict):
    """Upsert an item to a STAC API."""
    items_url = f"{endpoint}/collections/{collection_id}/items"
    response = requests.post(items_url, json=item.to_dict(), headers=headers)

    if response.status_code == 409:
        item_update_url = f"{items_url}/{item.id}"
        response = requests.put(item_update_url, json=item.to_dict(), headers=headers)
    elif response.status_code != 200:
        return f"Response from STAC API: {response.status_code}"
    if not response.ok:
        return f"Response from STAC API: {response.status_code}"


def delete_collection(endpoint: str, collection_id: str, headers: dict):
    """Upsert a collection to a STAC API."""
    collections_url = f"{endpoint}/collections/{collection_id}"
    return requests.delete(collections_url, headers=headers)


def download_model_from_stac_item(
    stac_item: pystac.Item,
    ras_directory: str,
    client: boto3.session.Session.client,
    bucket: str,
) -> str:
    """Download HEC-RAS model from stac href. Return the terrain_name."""
    # make RAS directory if it does not exists
    if not os.path.exists(ras_directory):
        os.makedirs(ras_directory)

    # download HEC-RAS model files
    for _, asset in stac_item.get_assets(role="hec-ras").items():
        s3_key = asset.extra_fields["s3_key"]

        file = os.path.join(ras_directory, Path(s3_key).name)
        client.download_file(bucket, s3_key, file)

    # download HEC-RAS topo files
    for _, asset in stac_item.get_assets(role="ras-topo").items():
        s3_key = asset.extra_fields["s3_key"]

        file = os.path.join(ras_directory, Path(s3_key).name)
        client.download_file(bucket, s3_key, file)

        if ".hdf" in Path(s3_key).name:
            terrain_name = Path(s3_key).name.rstrip(".hdf")

    return terrain_name



"""Conflate HEC-RAS Model."""

import json
import logging
import os
import traceback
from datetime import datetime
from urllib.parse import quote

import boto3
import pandas as pd
import pystac

import ripple1d
from ripple1d.conflate.plotter import plot_conflation_results
from ripple1d.conflate.rasfim import (
    RasFimConflater,
    nearest_line_to_point,
    ras_reaches_metadata,
    walk_network,
)
from ripple1d.ops.metrics import compute_conflation_metrics

logging.getLogger("fiona").setLevel(logging.ERROR)
logging.getLogger("botocore").setLevel(logging.ERROR)


def href_to_vsis(href: str, bucket: str) -> str:
    """Convert public aws href to a virtual ref for gdal read."""
    return href.replace(f"https://{bucket}.s3.amazonaws.com", f"/vsis3/{bucket}")


def s3_public_href(bucket: str, key: str) -> str:
    """Convert bucket and key to public href."""
    return f"https://{bucket}.s3.amazonaws.com/{quote(key)}"


def conflate_single_nwm_reach(rfc: RasFimConflater, nwm_reach_id: int):
    """Conflate a HEC-RAS model with a specific NWM reach."""
    nwm_reach_id_identified = False
    model_local_nwm_reaches = rfc.local_nwm_reaches()
    for river_reach_name in rfc.ras_river_reach_names:
        local_nwm_reaches = model_local_nwm_reaches.intersects(
            rfc.ras_xs_concave_hull(river_reach_name)["geometry"].iloc[0]
        )

        ras_start_point, ras_stop_point = rfc.ras_start_end_points(river_reach_name=river_reach_name)
        us_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_start_point)
        ds_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_stop_point)

        potential_reach_path = walk_network(local_nwm_reaches, us_most_reach_id, ds_most_reach_id)
        candidate_reaches = local_nwm_reaches.query(f"ID in {potential_reach_path}")
        if len(candidate_reaches.query(f"ID == {nwm_reach_id}")) == 1:
            nwm_reach_id_identified = True
            return ras_reaches_metadata(rfc, candidate_reaches[candidate_reaches["ID"] == nwm_reach_id])
    if not nwm_reach_id_identified:
        raise ValueError(f"nwm_reach_id {nwm_reach_id} not conflating to the ras model geometry.")


def conflate_model(source_model_directory: str, source_network: dict):
    """Conflate a HEC-RAS model with NWM reaches.

    source_network example:
    {
        "file_name": "nwm_flows_v3.parquet", // required
        "version": "2.1" // could be empty if not provided by the caller
        "type": "nwm_hydrofabric" // required
    }
    """
    try:
        nwm_pq_path = source_network["file_name"]
    except KeyError:
        raise KeyError(f"source_network must contain 'file_name', invalid parameters: {source_network}")

    if not source_network["type"] == "nwm_hydrofabric":
        raise ValueError(f"source_network type must be 'nwm_hydrofabric', invalid parameters: {source_network}")

    version = source_network.get("version", "")

    rfc = RasFimConflater(nwm_pq_path, source_model_directory)
    metadata = {"reaches": {}}
    # get nwm reaches that intersect the convex hull of the ras model
    model_local_nwm_reaches = rfc.local_nwm_reaches()
    for river_reach_name in rfc.ras_river_reach_names:

        # get the concave hull of the this river reach
        concave_hull = rfc.ras_xs_concave_hull(river_reach_name)

        # get the nwm reaches that intersect the concave hull of the cross section of this river reach using the convex hull for this model
        local_nwm_reaches = model_local_nwm_reaches[
            model_local_nwm_reaches.intersects(concave_hull["geometry"].iloc[0])
        ]

        # get the start and end points of the river reach
        ras_start_point, ras_stop_point = rfc.ras_start_end_points(river_reach_name=river_reach_name, clip_to_xs=True)

        # get the nearest upstream and downstream nwm reaches to the start and end points of the river reach
        if len(rfc.ras_river_reach_names) == 1:
            # if there are multiple river reaches, don't raise error; quitely continue to the next river reach
            us_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_start_point)
            ds_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_stop_point)
        else:
            # if this is the only river reach, raise error
            try:
                us_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_start_point)
                ds_most_reach_id = nearest_line_to_point(local_nwm_reaches, ras_stop_point)
            except ValueError as e:
                logging.error(f"Error: {e}")
                continue

        logging.info(
            f"{river_reach_name} | us_most_reach_id ={us_most_reach_id} and ds_most_reach_id = {ds_most_reach_id}"
        )

        # walk network to get the potential reach ids
        potential_reach_path = walk_network(local_nwm_reaches, us_most_reach_id, ds_most_reach_id)
        potential_reach_path = list(set(potential_reach_path) - set(metadata.keys()))

        # get gdf of the candidate reaches
        candidate_reaches = local_nwm_reaches.query(f"ID in {potential_reach_path}")

        metadata["reaches"].update(ras_reaches_metadata(rfc, candidate_reaches))

    if not conflated(metadata):
        return "no reaches conflated"

    ids = list(metadata["reaches"].keys())
    fim_stream = rfc.local_nwm_reaches()[rfc.local_nwm_reaches()["ID"].isin(ids)]
    conflation_png = f"{rfc.ras_gpkg.replace('.gpkg','.conflation.png')}"

    plot_conflation_results(
        rfc,
        fim_stream,
        conflation_png,
        limit_plot_to_nearby_reaches=True,
    )

    logging.info(f"Conflation results: {metadata}")
    conflation_file = f"{rfc.ras_gpkg.replace('.gpkg','.conflation.json')}"

    metadata["metadata"] = {}
    metadata["metadata"]["source_network"] = source_network.copy()
    metadata["metadata"]["source_network"]["file_name"] = os.path.basename(nwm_pq_path)
    metadata["metadata"]["conflation_png"] = os.path.basename(conflation_png)
    metadata["metadata"]["conflation_ripple1d_version"] = ripple1d.__version__
    metadata["metadata"]["metrics_ripple1d_version"] = ripple1d.__version__
    metadata["metadata"]["source_ras_model"] = {
        "stac_api": rfc.stac_api,
        "stac_collection_id": rfc.stac_collection_id,
        "stac_item_id": rfc.stac_item_id,
    }
    metadata["metadata"]["source_ras_model"]["source_ras_files"] = {
        "geometry": rfc.primary_geom_file,
        "forcing": rfc.primary_flow_file,
        "project-file": rfc.ras_project_file,
        "plan": rfc.primary_plan_file,
    }

    with open(conflation_file, "w") as f:
        f.write(json.dumps(metadata, indent=4))

    try:
        compute_conflation_metrics(source_model_directory, source_network)
    except Exception as e:
        logging.error(f"Error: {e}")
        logging.error(f"traceback: {traceback.format_exc()}")

    return conflation_file


def conflated(metadata: dict) -> bool:
    """Determine if any reaches conflated."""
    count = 0
    for reach_data in metadata["reaches"].values():
        if not reach_data["eclipsed"]:
            count += 1
    if count == 0:
        return False
    else:
        return True


# def conflate_s3_model(
#     item: pystac.Item, client: boto3.client, bucket: str, stac_item_s3_key: str, rfc: RasFimConflater
# ):
#     """Conflate a model from s3."""
#     # build conflation key and href
#     nwm_conflation_key = stac_item_s3_key.replace(".json", ".conflation.json")
#     nwm_conflation_href = s3_public_href(bucket, nwm_conflation_key)

#     # conflate the mip ras model to nwm reaches
#     conflation_results = conflate(rfc)

#     # write conflation results to s3
#     client.put_object(
#         Body=json.dumps(conflation_results).encode(),
#         Bucket=bucket,
#         Key=nwm_conflation_key,
#     )

#     # Add ripple parameters asset to item
#     item.add_asset(
#         "NWM_Conflation",
#         pystac.Asset(
#             nwm_conflation_href,
#             title="NWM_Conflation",
#             roles=[pystac.MediaType.JSON, "nwm-conflation"],
#             extra_fields={
#                 "software": f"ripple1d {ripple1d.__version__}",
#                 "date_created": datetime.now().isoformat(),
#             },
#         ),
#     )

#     limit_plot = True

#     conflation_thumbnail_key = nwm_conflation_key.replace("json", "png")
#     conflation_thumbnail_href = s3_public_href(bucket, conflation_thumbnail_key)

#     ids = list(conflation_results["reaches"].keys())

#     fim_stream = rfc.local_nwm_reaches()[rfc.local_nwm_reaches()["ID"].isin(ids)]

#     plot_conflation_results(
#         rfc,
#         fim_stream,
#         conflation_thumbnail_key,
#         bucket=bucket,
#         s3_client=client,
#         limit_plot_to_nearby_reaches=limit_plot,
#     )
#     # Add thumbnail asset to item
#     item.add_asset(
#         "ThumbnailConflationResults",
#         pystac.Asset(
#             conflation_thumbnail_href,
#             title="ThumbnailConflationResults",
#             roles=[pystac.MediaType.PNG, "thumbnail"],
#             extra_fields={
#                 "software": f"ripple1d {ripple1d.__version__}",
#                 "date_created": datetime.now().isoformat(),
#             },
#             description="""PNG of NWM conflation results with OpenStreetMap basemap.""",
#         ),
#     )

#     client.put_object(Body=json.dumps(item.to_dict()).encode(), Bucket=bucket, Key=stac_item_s3_key)



"""Generic utilities supporting the huey + Flask REST API."""

import inspect
import logging
import os
import sysconfig
import traceback
import typing
from functools import wraps


def tracerbacker(func: typing.Callable) -> tuple[typing.Any, str | None, str | None]:
    """Call the wrapped function and return 3 items: a jsonified version of the function's returned value, an error message, and a traceback string.

    If the wrapped function *does not* raise an exception:
        Response will be jsonified version of the wrapped function's returned value.
        Error message will be None
        Traceback will be None

    If the wrapped function *does* raise an exception:
        Response will be None
        Error message will be str(exception)
        Traceback will be a string of the exception traceback

    The error message and traceback string will each be an empty string if the function did not raise an exception.
    """

    @wraps(func)  # this ensures the func name does not change in huey
    def wrapper(*args, **kwargs):
        try:
            return_value = func(*args, **kwargs)
        except Exception as e:
            return_value = None
            err_msg = str(e)
            tb = traceback.format_exc()
            logging.error(err_msg)
        else:
            err_msg = None
            tb = None
        return {"val": return_value, "err": err_msg, "tb": tb}

    return wrapper


def get_unexpected_and_missing_args(func: typing.Callable, kwargs_provided: set) -> tuple[list, list]:
    """Inspect func and evaluate whether the provided kwargs will satisfy it, considering its required & optional args.

    Return a list of unexpected kwargs (provided, but not called for),
    and return a list of missing kwargs (required but not provided).
    WARNING: assumes the func does not use *args or **kwargs features.
    """
    argspec = inspect.getfullargspec(func)
    unexpected = kwargs_provided - set(argspec.args)
    if argspec.defaults:
        required = set(argspec.args[: -len(argspec.defaults)])
    else:
        required = set(argspec.args)
    missing = required - kwargs_provided
    return sorted(unexpected), sorted(missing)


def find_huey_consumer():
    """Find the huey_consumer.py file in the installed packages and return its path."""
    site_packages_path = sysconfig.get_paths()["purelib"]

    for root, dirs, files in os.walk(site_packages_path):
        if "huey_consumer.py" in files:
            huey_consumer_path = os.path.join(root, "huey_consumer.py")
            print(f"Found huey_consumer.py at: {huey_consumer_path}")
            return huey_consumer_path

    print("huey_consumer.py not found in installed packages.")
    return None



"""Plotting functions for geometry data."""

import logging
from io import BytesIO
from pathlib import Path

import geopandas as gpd
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
import shapely
from boto3 import Session

from .rasfim import RasFimConflater


def plot_conflation_results(
    rfc: RasFimConflater,
    fim_stream: gpd.GeoDataFrame,
    key: str,
    bucket: str = None,
    s3_client: Session.client = None,
    limit_plot_to_nearby_reaches: bool = True,
):
    """Create/write png to s3. The png contains RAS centerline and cross sections and nearby NWM branches."""
    _, ax = plt.subplots(figsize=(10, 10))

    # Plot the centerline and cross-sections first
    rfc.ras_centerlines.plot(ax=ax, color="black", label="RAS Centerline", alpha=0.5, linestyle="dashed")
    rfc.ras_xs.plot(ax=ax, color="green", label="RAS XS", markersize=2, alpha=0.2)
    if rfc.ras_structures is not None:
        rfc.ras_structures.plot(ax=ax, color="grey", label="RAS Structure", markersize=2, alpha=0.2)

    # Get the current axis limits and create a rectangle geometry
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    bounds = shapely.geometry.box(xlim[0], ylim[0], xlim[1], ylim[1])

    zoom_factor = 1  # Adjust this value to change the zoom level

    # Calculate the range of x and y
    x_range = bounds.bounds[2] - bounds.bounds[0]
    y_range = bounds.bounds[3] - bounds.bounds[1]

    min_x = bounds.bounds[0] - x_range * (zoom_factor - 1) / 2
    max_x = bounds.bounds[2] + x_range * (zoom_factor - 1) / 2
    min_y = bounds.bounds[1] - y_range * (zoom_factor - 1) / 2
    max_y = bounds.bounds[3] + y_range * (zoom_factor - 1) / 2

    adjusted_bounds = shapely.geometry.box(min_x, min_y, max_x, max_y)
    # Add a patch for the ras_centerline
    patches = [mpatches.Patch(color="black", label="RAS Centerline", linestyle="dashed")]

    # Add a patch for nearby reaches
    patches.append(mpatches.Patch(color="blue", label="Nearby NWM reaches", alpha=0.3))

    # Plot the reaches that fall within the axis limits
    rfc.local_nwm_reaches().plot(ax=ax, color="blue", linewidth=1, alpha=0.3)

    if limit_plot_to_nearby_reaches:
        # Create a colormap that maps each reach_id to a color
        unique_reach_ids = fim_stream["ID"].unique()
        colors = plt.cm.viridis(np.linspace(0, 1, len(unique_reach_ids)))
        colormap = dict(zip(unique_reach_ids, colors))

        # Plot the fim_stream using the colormap
        fim_stream["color"] = fim_stream["ID"].map(colormap)
        fim_stream.plot(color=fim_stream["color"], ax=ax, linewidth=2, alpha=0.8)
        # Create a custom legend using the colormap
        patches.extend(
            [mpatches.Patch(color=colormap[reach_id], label=f"reach {reach_id}") for reach_id in unique_reach_ids]
        )

    zoom_factor = 3  # Adjust this value to change the zoom level

    # Calculate the range of x and y
    x_range = bounds.bounds[2] - bounds.bounds[0]
    y_range = bounds.bounds[3] - bounds.bounds[1]

    # Set the axis limits to the bounds, expanded by the zoom factor
    ax.set_xlim(
        min_x,
        max_x,
    )
    ax.set_ylim(
        min_y,
        max_y,
    )

    ax.legend(handles=patches, handleheight=0.005)
    ax.set_xticks([])
    ax.set_yticks([])

    plt.tight_layout()

    if s3_client:
        buf = BytesIO()
        plt.savefig(buf, format="png")
        buf.seek(0)
        s3_client.put_object(Bucket=bucket, Key=key, Body=buf, ContentType="image/png")
    else:
        output_png = str(Path(rfc.source_model_directory) / Path(key).name)
        plt.savefig(output_png, format="png")

    plt.close()



"""Utils for working with s3."""

import json
import logging
import os
import pathlib
import posixpath
import re
import traceback
from datetime import datetime, timezone
from urllib.parse import urlparse

import boto3
import boto3.session
import botocore
import pystac
from mypy_boto3_s3.service_resource import ObjectSummary

logging.getLogger("boto3").setLevel(logging.WARNING)
logging.getLogger("botocore").setLevel(logging.WARNING)


def str_from_s3(ras_text_file_path: str, client, bucket) -> str:
    """Read a text file from s3 and return its contents as a string."""
    logging.debug(f"reading: {ras_text_file_path}")
    response = client.get_object(Bucket=bucket, Key=ras_text_file_path)
    return response["Body"].read().decode("utf-8")


def list_keys(s3_client: boto3.Session.client, bucket: str, prefix: str, suffix: str = "") -> list:
    """List all keys in an S3 bucket with a given prefix and suffix."""
    keys = []
    kwargs = {"Bucket": bucket, "Prefix": prefix}
    while True:
        resp = s3_client.list_objects_v2(**kwargs)
        keys += [obj["Key"] for obj in resp["Contents"] if obj["Key"].endswith(suffix)]
        try:
            kwargs["ContinuationToken"] = resp["NextContinuationToken"]
        except KeyError:
            break
    return keys


def list_keys_regex(s3_client: boto3.Session.client, bucket: str, prefix_includes: str, suffix="") -> list:
    """List all keys in an S3 bucket with a given prefix and suffix."""
    keys = []
    kwargs = {"Bucket": bucket, "Prefix": prefix_includes}
    prefix_pattern = re.compile(prefix_includes.replace("*", ".*"))
    while True:
        resp = s3_client.list_objects_v2(**kwargs)
        keys += [
            obj["Key"] for obj in resp["Contents"] if prefix_pattern.match(obj["Key"]) and obj["Key"].endswith(suffix)
        ]
        try:
            kwargs["ContinuationToken"] = resp["NextContinuationToken"]
        except KeyError:
            break
    return keys


def init_s3_resources() -> tuple:
    """Establish a boto3 (AWS) session and return the session, S3 client, and S3 resource handles."""
    # Instantitate S3 resources
    session = boto3.Session(
        aws_access_key_id=os.environ.get("AWS_ACCESS_KEY_ID"),
        aws_secret_access_key=os.environ.get("AWS_SECRET_ACCESS_KEY"),
    )

    s3_client = session.client("s3")
    s3_resource = session.resource("s3")
    return session, s3_client, s3_resource


def check_s3_key_exists(bucket: str, key: str) -> bool:
    """Check if an object with the given key exists in the specified S3 bucket."""
    s3 = boto3.client("s3")
    try:
        s3.head_object(Bucket=bucket, Key=key)
        return True
    except Exception:
        return False


def read_json_from_s3(bucket: str, key: str) -> dict:
    """Read a JSON file from an S3 bucket and return its contents as a dictionary."""
    s3_client = boto3.client("s3")
    response = s3_client.get_object(Bucket=bucket, Key=key)
    file_content = response["Body"].read().decode("utf-8")
    json_content = json.loads(file_content)
    return json_content


def get_basic_object_metadata(obj: ObjectSummary) -> dict:
    """
    Retrieve basic metadata of an AWS S3 object.

    Parameters
    ----------
        obj (ObjectSummary): The AWS S3 object.

    Returns
    -------
        dict: A dictionary with the size, ETag, last modified date, storage platform, region, and storage tier of the object.
    """
    try:
        _ = obj.load()
        return {
            "file:size": obj.content_length,
            "e_tag": obj.e_tag.strip('"'),
            "last_modified": obj.last_modified.isoformat(),
            "storage:platform": "AWS",
            "storage:region": obj.meta.client.meta.region_name,
            "storage:tier": obj.storage_class,
        }
    except botocore.exceptions.ClientError:
        raise KeyError(f"Unable to access {obj.key} check that key exists and you have access")


def copy_item_to_s3(item: pystac.Item, s3_key: str, s3client: botocore.client.BaseClient):
    """
    Copy an item to an AWS S3 bucket.

    Parameters
    ----------
        item: The item to copy. It must have a `to_dict` method that returns a dictionary representation of it.
        s3_key (str): The file path in the S3 bucket to copy the item to.

    The function performs the following steps:
        1. Initializes a boto3 S3 client and splits the s3_key into the bucket name and the key.
        2. Converts the item to a dictionary, serializes it to a JSON string, and encodes it to bytes.
        3. Puts the encoded JSON string to the specified file path in the S3 bucket.
    """
    # s3 = boto3.client("s3")
    bucket, key = split_s3_key(s3_key)
    s3client.put_object(Body=json.dumps(item.to_dict()).encode(), Bucket=bucket, Key=key)


def split_s3_key(s3_key: str) -> tuple[str, str]:
    """
    Split an S3 key into the bucket name and the key.

    Parameters
    ----------
        s3_key (str): The S3 key to split. It should be in the format 's3://bucket/key'.

    Returns
    -------
        tuple: A tuple containing the bucket name and the key. If the S3 key does not contain a key, the second element
          of the tuple will be None.

    The function performs the following steps:
        1. Removes the 's3://' prefix from the S3 key.
        2. Splits the remaining string on the first '/' character.
        3. Returns the first part as the bucket name and the second part as the key. If there is no '/', the key will
          be None.
    """
    parts = s3_key.replace("s3://", "").split("/", 1)
    bucket = parts[0]
    key = parts[1] if len(parts) > 1 else None
    return bucket, key


def s3_key_public_url_converter(url: str, dev_mode: bool = False) -> str:
    """
    Convert an S3 URL to an HTTPS URL and vice versa.

    Args:
    ----------
        url (str): The URL to convert. It should be in the format 's3://bucket/' or 'https://bucket.s3.amazonaws.com/'.
        dev_mode (bool): A flag indicating whether the function should use the Minio endpoint for S3 URL conversion.
    Return:
    -------
        str: The converted URL. If the input URL is an S3 URL, the function returns an HTTPS URL. If the input URL is
        an HTTPS URL, the function returns an S3 URL.

    The function performs the following steps:
        1. Checks if the input URL is an S3 URL or an HTTPS URL.
        2. If the input URL is an S3 URL, it converts it to an HTTPS URL.
        3. If the input URL is an HTTPS URL, it converts it to an S3 URL.
    """
    if url.startswith("s3"):
        bucket = url.replace("s3://", "").split("/")[0]
        key = url.replace(f"s3://{bucket}", "")[1:]
        if dev_mode:
            logging.info(f"dev_mode | using minio endpoint for s3 url conversion: {url}")
            return f"{os.environ.get('MINIO_S3_ENDPOINT')}/{bucket}/{key}"
        else:
            return f"https://{bucket}.s3.amazonaws.com/{key}"

    elif url.startswith("http"):
        if dev_mode:
            logging.info(f"dev_mode | using minio endpoint for s3 url conversion: {url}")
            bucket = url.replace(os.environ.get("MINIO_S3_ENDPOINT"), "").split("/")[0]
            key = url.replace(os.environ.get("MINIO_S3_ENDPOINT"), "")
        else:
            bucket = url.replace("https://", "").split(".s3.amazonaws.com")[0]
            key = url.replace(f"https://{bucket}.s3.amazonaws.com/", "")

        return f"s3://{bucket}/{key}"

    else:
        raise ValueError(f"Invalid URL format: {url}")


def verify_safe_prefix(s3_key: str):
    """
    TODO: discuss this with the team. Would like some safety mechanism to ensure that the S3 key is limited to certain prefixes.

    Should there be some restriction where these files can be written?
    """
    parts = s3_key.split("/")
    logging.debug(f"parts of the s3_key: {parts}")
    if parts[3] != "stac":
        raise ValueError(f"prefix must begin with stac/, user provided {s3_key} needs to be corrected")


def s3_upload_dir_recursively(local_src_dir: str, tgt_dir: str, s3_client: botocore.client.BaseClient) -> None:
    """Copy all files from a local directory. tgt_dir can be local or a s3:// prefix."""
    assert tgt_dir.startswith("s3://")
    pathmod = posixpath
    if not os.path.isdir(local_src_dir):
        raise NotADirectoryError(local_src_dir)
    for root, _, files in os.walk(local_src_dir):
        rel_root = os.path.relpath(root, start=local_src_dir)
        if os.path is not posixpath:
            # copying to s3 (posix system), but running in Windows
            rel_root = pathlib.PurePath(rel_root).as_posix()
        if rel_root == ".":
            rel_root = ""
        for fn in files:
            src_file = os.path.join(root, fn)
            tgt_file = pathmod.join(tgt_dir, rel_root, fn)
            logging.debug(f"Uploading: {src_file} -> {tgt_file}")
            bucket_name, key = extract_bucketname_and_keyname(s3path=tgt_file)
            s3_client.upload_file(
                Filename=src_file,
                Bucket=bucket_name,
                Key=key,
            )


def s3_delete_dir_recursively(s3_dir: str, s3_resource: boto3.resources.factory.ServiceResource) -> None:
    """Delete a s3:// directory and its contents recursively. OK if dir does not exist."""
    logging.debug(f"Deleting directory if exists: {s3_dir}")
    if not s3_dir.startswith("s3://"):
        raise ValueError(f"Expected s3_dir to start with s3://, but got: {s3_dir}")
    bucket, key = extract_bucketname_and_keyname(s3path=s3_dir)
    if not key.strip():
        raise ValueError(f"s3 path too short: {s3_dir}")
    if len(key.split("/")) < 3:
        raise ValueError(f"s3 path too short: {s3_dir}")
    if not key.endswith("/"):
        key += "/"
    bucket_handle = s3_resource.Bucket(bucket)
    bucket_handle.objects.filter(Prefix=key).delete()


def extract_bucketname_and_keyname(s3path: str) -> tuple[str, str]:
    """Parse the provided s3:// object path and return its bucket name and key."""
    if not s3path.startswith("s3://"):
        raise ValueError(f"s3path does not start with s3://: {s3path}")
    bucket, _, key = s3path[5:].partition("/")
    return bucket, key


def s3_upload_status_file(
    stac_href: str,
    s3_bucket: str,
    s3_client: botocore.client.BaseClient,
    e: Exception | None,
):
    """
    Upload a status file to s3.

    If e is a Python exception, then upload a 'fail' json file to the href's standard
    output location on s3.  If e is None, then upload a 'succeed' json file.  Either file
    will have key "time" indicating the time that the file was uploaded.  A 'fail' file will
    also have keys "err" and "traceback" containing the exception as a string and the Python
    traceback of the exception, respectively.
    """
    s3_output_key_succeed, s3_output_key_fail = s3_get_ripple1d_status_file_key_names(stac_href, s3_bucket, s3_client)

    time_now_str = datetime.now(tz=timezone.utc).isoformat()
    if e is None:
        s3_output_key = s3_output_key_succeed
        body = {"time": time_now_str}
    elif isinstance(e, Exception):
        s3_output_key = s3_output_key_fail
        body = {
            "time": time_now_str,
            "err": str(e),
            "traceback": "".join(traceback.format_tb(e.__traceback__)),
        }
    else:
        raise TypeError(f"For e, expected None or type Exception, but got type: {type(e)}")

    logging.debug(f"Deleting if exists: {s3_output_key_succeed}")
    s3_client.delete_object(Bucket=s3_bucket, Key=s3_output_key_succeed)
    logging.debug(f"Deleting if exists: {s3_output_key_fail}")
    s3_client.delete_object(Bucket=s3_bucket, Key=s3_output_key_fail)

    body_str = json.dumps(body, indent=2)
    logging.debug(f"Writing: {s3_output_key} with body: {body_str}")
    s3_client.put_object(
        Body=body_str,
        Bucket=s3_bucket,
        Key=s3_output_key,
        ContentType="application/json",
    )


def s3_ripple1d_status_succeed_file_exists(
    stac_href: str, s3_bucket: str, s3_client: botocore.client.BaseClient
) -> bool:
    """Check if the standard ripple1d succeed sentinel file exists.  If it does, return True, otherwise return False."""
    s3_output_key_succeed, _ = s3_get_ripple1d_status_file_key_names(stac_href, s3_bucket, s3_client)
    logging.debug(f"Checking if s3 file exists: s3://{s3_bucket}/{s3_output_key_succeed}")
    try:
        s3_client.head_object(Bucket=s3_bucket, Key=s3_output_key_succeed)
    except botocore.exceptions.ClientError as e:
        if "Not Found" in str(e):
            return False  # typical ClientError when the object does not exist
        else:
            raise  # unexpected ClientError
    return True


def s3_get_ripple1d_status_file_key_names(
    stac_href: str, s3_bucket: str, s3_client: botocore.client.BaseClient
) -> tuple[str, str]:
    """Return two S3 key paths, the first to a succeed sentinel file, the second t oa failure sentinel file.

    This function does not check if the keys exist.
    """
    _, s3_output_dir_key = extract_bucketname_and_keyname(s3_get_output_s3path(s3_bucket, stac_href))
    s3_output_key_succeed = posixpath.join(s3_output_dir_key, "ripple1d-succeed.json")
    s3_output_key_fail = posixpath.join(s3_output_dir_key, "ripple1d-fail.json")
    return s3_output_key_succeed, s3_output_key_fail


def s3_get_output_s3path(s3_bucket: str, stac_href: str) -> str:
    """Return the s3 path to the output directory for the given stac_href."""
    return f"s3://{s3_bucket}/mip/dev/ripple1d/output{urlparse(stac_href).path}/"



"""Run HEC-RAS models."""

import json
import logging
import os
from dataclasses import asdict
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd

from ripple1d.consts import DEFAULT_EPSG, MIN_FLOW
from ripple1d.data_model import FlowChangeLocation, NwmReachModel
from ripple1d.ras import RasManager


def create_model_run_normal_depth(
    submodel_directory: str,
    plan_suffix: str,
    num_of_discharges_for_initial_normal_depth_runs: int = 10,
    ras_version: str = "631",
    show_ras: bool = False,
):
    """Write and compute initial normal depth runs to develop initial rating curves."""
    nwm_rm = NwmReachModel(submodel_directory)

    if not nwm_rm.file_exists(nwm_rm.conflation_file):
        raise FileNotFoundError(f"cannot find conflation file {nwm_rm.conflation_file}, please ensure file exists")

    if not nwm_rm.file_exists(nwm_rm.ras_gpkg_file):
        raise FileNotFoundError(f"cannot find ras_gpkg_file file {nwm_rm.ras_gpkg_file}, please ensure file exists")

    if nwm_rm.ripple1d_parameters["us_xs"]["xs_id"] == "-9999":
        logging.warning(f"skipping {nwm_rm.model_name}; no cross sections conflated.")
    else:
        logging.info(f"Working on initial normal depth run for nwm_id: {nwm_rm.model_name}")

        # create new ras manager class
        rm = RasManager.from_gpkg(nwm_rm.ras_project_file, nwm_rm.model_name, nwm_rm.ras_gpkg_file, ras_version)

        # increment flows based on min and max flows specified in conflation parameters
        initial_flows = np.linspace(
            max([nwm_rm.ripple1d_parameters["low_flow"], MIN_FLOW]),
            nwm_rm.ripple1d_parameters["high_flow"],
            num_of_discharges_for_initial_normal_depth_runs,
        ).astype(int)

        # # write and compute initial normal depth runs to develop rating curves
        fcl = FlowChangeLocation(
            nwm_rm.model_name,
            nwm_rm.model_name,
            rm.geoms[nwm_rm.model_name].rivers[nwm_rm.model_name][nwm_rm.model_name].us_xs.river_station,
            initial_flows.tolist(),
        )

        rm.normal_depth_run(
            f"{nwm_rm.model_name}_{plan_suffix}",
            nwm_rm.model_name,
            [fcl],
            initial_flows.astype(str),
            write_depth_grids=False,
            show_ras=show_ras,
            run_ras=True,
        )
    return {f"{nwm_rm.model_name}_{plan_suffix}": asdict(fcl)}


def run_incremental_normal_depth(
    submodel_directory: str,
    plan_suffix: str,
    ras_version: str = "631",
    depth_increment=0.5,
    write_depth_grids: str = True,
    show_ras: bool = False,
):
    """Write and compute incremental normal depth runs to develop rating curves and depth grids."""
    nwm_rm = NwmReachModel(submodel_directory)

    if not nwm_rm.file_exists(nwm_rm.conflation_file):
        raise FileNotFoundError(f"cannot find conflation file {nwm_rm.conflation_file}, please ensure file exists")

    if not nwm_rm.file_exists(nwm_rm.ras_gpkg_file):
        raise FileNotFoundError(f"cannot find ras_gpkg_file file {nwm_rm.ras_gpkg_file}, please ensure file exists")

    logging.info(f"Working on normal depth run for nwm_id: {nwm_rm.model_name}")
    if nwm_rm.ripple1d_parameters["us_xs"]["xs_id"] == "-9999":
        logging.warning(f"skipping {nwm_rm.model_name}; no cross sections conflated.")

    rm = RasManager(
        nwm_rm.ras_project_file,
        version=ras_version,
        terrain_path=nwm_rm.ras_terrain_hdf,
        crs=nwm_rm.ripple1d_parameters["crs"],
    )

    # determine flow increments
    flows, _, _ = determine_flow_increments(
        rm,
        [f"{nwm_rm.model_name}_ind"],
        nwm_rm.model_name,
        nwm_rm.model_name,
        nwm_rm.model_name,
        depth_increment=depth_increment,
    )

    fcl = FlowChangeLocation(
        nwm_rm.model_name,
        nwm_rm.model_name,
        rm.geoms[nwm_rm.model_name].rivers[nwm_rm.model_name][nwm_rm.model_name].us_xs.river_station,
        flows.tolist(),
    )
    # write and compute flow/plans for normal_depth run
    rm.normal_depth_run(
        f"{nwm_rm.model_name}_{plan_suffix}",
        nwm_rm.model_name,
        [fcl],
        flows.astype(str),
        write_depth_grids=True,
        show_ras=show_ras,
        run_ras=True,
    )
    return {f"{nwm_rm.model_name}_{plan_suffix}": asdict(fcl)}


def run_known_wse(
    submodel_directory: str,
    plan_suffix: str,
    min_elevation: float,
    max_elevation: float,
    depth_increment=2,
    ras_version: str = "631",
    write_depth_grids: str = True,
    show_ras: bool = False,
):
    """Write and compute known water surface elevation runs to develop rating curves and depth grids."""
    nwm_rm = NwmReachModel(submodel_directory)

    if not nwm_rm.file_exists(nwm_rm.conflation_file):
        raise FileNotFoundError(f"cannot find conflation file {nwm_rm.conflation_file}, please ensure file exists")

    if not nwm_rm.file_exists(nwm_rm.ras_gpkg_file):
        raise FileNotFoundError(f"cannot find ras_gpkg_file file {nwm_rm.ras_gpkg_file}, please ensure file exists")

    logging.info(f"Working on known water surface elevation run for nwm_id: {nwm_rm.model_name}")

    start_elevation = np.floor(min_elevation * 2) / 2  # round down to nearest .0 or .5
    known_water_surface_elevations = np.arange(start_elevation, max_elevation + depth_increment, depth_increment)

    crs = gpd.read_file(nwm_rm.ras_gpkg_file, layer="XS").crs

    # write and compute flow/plans for known water surface elevation runs
    rm = RasManager(nwm_rm.ras_project_file, version=ras_version, terrain_path=nwm_rm.ras_terrain_hdf, crs=crs)

    # get resulting depths from the second normal depth runs_nd
    rm.plan = rm.plans[f"{nwm_rm.model_name}_nd"]
    ds_flows, ds_depths, _ = get_flow_depth_arrays(
        rm,
        nwm_rm.model_name,
        nwm_rm.model_name,
        rm.geoms[nwm_rm.model_name].rivers[nwm_rm.model_name][nwm_rm.model_name].ds_xs.river_station,
        rm.geoms[nwm_rm.model_name].rivers[nwm_rm.model_name][nwm_rm.model_name].ds_xs.thalweg,
    )

    known_depths = (
        known_water_surface_elevations
        - rm.geoms[nwm_rm.model_name].rivers[nwm_rm.model_name][nwm_rm.model_name].ds_xs.thalweg
    )

    # filter known water surface elevations less than depths resulting from the second normal depth run
    depths, flows, wses = create_flow_depth_combinations(
        known_depths,
        known_water_surface_elevations,
        ds_flows,
        ds_depths,
    )

    if not flows:
        logging.warning(
            f"No controling known water surface elevations were identified for {nwm_rm.model_name}; i.e., the depth of flooding\
 for the normal depth run for a given flow was alway higher than the known water surface elevations of the downstream reach"
        )
    else:
        rm.kwses_run(
            f"{nwm_rm.model_name}_{plan_suffix}",
            nwm_rm.model_name,
            depths,
            wses,
            flows,
            nwm_rm.model_name,
            nwm_rm.model_name,
            rm.geoms[nwm_rm.model_name].rivers[nwm_rm.model_name][nwm_rm.model_name].us_xs.river_station,
            write_depth_grids=write_depth_grids,
            show_ras=show_ras,
            run_ras=True,
        )
    return {f"{nwm_rm.model_name}_{plan_suffix}": {"kwse": known_water_surface_elevations.tolist()}}


def get_flow_depth_arrays(
    rm: RasManager, river: str, reach: str, river_station: str, thalweg: float
) -> tuple[pd.Series]:
    """Create new flow, depth,wse arrays from rating curve-plans results."""
    # read in flow/wse
    wses, flows = rm.plan.read_rating_curves()

    # get the river_reach_rs for the cross section representing the upstream end of this reach
    river_reach_rs = f"{river} {reach} {str(river_station)}"

    wse = wses.loc[river_reach_rs, :]
    flow = flows.loc[river_reach_rs, :]
    df = pd.DataFrame({"wse": wse.astype(int), "flow": flow.round(1)}).drop_duplicates()

    # convert wse to depth
    depth = df["wse"] - thalweg

    return (df["flow"], depth, df["wse"])


def determine_flow_increments(
    rm: RasManager,
    plan_names: str,
    river: str,
    reach: str,
    nwm_id: str,
    depth_increment: float = 0.5,
) -> tuple[np.array]:
    """Detemine flow increments corresponding to 0.5 ft depth increments using the rating-curve-run results."""
    flows, depths = [], []
    for plan_name in plan_names:
        rm.plan = rm.plans[plan_name]

        river_station = rm.geoms[nwm_id].rivers[nwm_id][nwm_id].us_xs.river_station
        thalweg = rm.geoms[nwm_id].rivers[nwm_id][nwm_id].us_xs.thalweg

        # get new flow/depth for current branch
        flow, depth, _ = get_flow_depth_arrays(rm, river, reach, river_station, thalweg)
        flows.append(np.array(flow))
        depths.append(np.array(depth))
    # get new flow/depth incremented every x ft
    new_depths, new_flows = create_flow_depth_array(np.concatenate(flows), np.concatenate(depths), depth_increment)

    new_wse = new_depths + thalweg  # [i + thalweg for i in new_depths]

    return new_flows.astype(int), new_depths, new_wse


def create_flow_depth_combinations(
    ds_depths: list, ds_wses: list, input_flows: np.array, min_depths: pd.Series
) -> tuple:
    """
    Create flow-depth-wse combinations.

    Args:
        ds_depths (list): downstream depths
        ds_wses (list): downstream water surface elevations
        input_flows (np.array): Flows to create profiles names from. Combine with incremental depths
            of the downstream cross section of the reach
        min_depths (pd.Series): minimum depth to be included. (typically derived from a previous noraml depth run)

    Returns
    -------
        tuple: tuple of depths, flows, and wses
    """
    depths, flows, wses = [], [], []
    for wse, depth in zip(ds_wses, ds_depths):
        for flow in input_flows:
            if depth >= min_depths.loc[str(int(flow))]:
                depths.append(round(depth, 1))
                flows.append(int(max([flow, MIN_FLOW])))
                wses.append(round(wse, 1))
    return (depths, flows, wses)


def get_kwse_from_ds_model(ds_nwm_id: str, ds_nwm_ras_project_file: str, plan_names: str) -> tuple[float]:
    """Get the kwse values from the downstream model."""
    rm = RasManager(ds_nwm_ras_project_file, crs=DEFAULT_EPSG)
    wses = []
    for plan_name in plan_names:
        if plan_name not in rm.plans.keys():
            logging.warning(f"{plan_name} is not an existing plan in the specified HEC-RAS model")
            return np.array([])

        rm.plan = rm.plans[plan_name]

        river_reach_rs = rm.plan.geom.rivers[ds_nwm_id][ds_nwm_id].us_xs.river_reach_rs
        thalweg = rm.plan.geom.rivers[ds_nwm_id][ds_nwm_id].us_xs.thalweg

        wse, _ = rm.plan.read_rating_curves()

        wses.append(wse.loc[river_reach_rs, :])

        df = pd.concat(wses)
    return df.min(), df.max()


def establish_order_of_nwm_ids(conflation_parameters: dict) -> list[str]:
    """Establish the order of NWM IDs based on the cross section IDs."""
    order = []
    for id, data in conflation_parameters.items():
        if conflation_parameters[id]["us_xs"]["xs_id"] == "-9999":
            logging.warning(f"skipping {id}; no cross sections conflated.")
        else:
            order.append((float(data["us_xs"]["xs_id"]), id))
    order.sort()
    return [i[1] for i in order]


def create_flow_depth_array(flow: list[float], depth: list[float], increment: float = 0.5) -> tuple[np.array]:
    """Interpolate flow values to a new depth array with a specified increment."""
    min_depth = np.min(depth)
    max_depth = np.max(depth)
    start_depth = np.floor(min_depth * 2) / 2  # round down to nearest .0 or .5
    new_depth = np.arange(start_depth, max_depth + increment, increment)
    new_flow = np.interp(new_depth, np.sort(depth), np.sort(flow))

    return new_depth, new_flow



"""Manage Flask and Huey."""

import argparse
import json
import os
import shutil
import subprocess
import sys
from datetime import datetime


class RippleManager:
    """Launch Flask and Huey in separate terminals and manage their lifecycles."""

    def __init__(self, args):
        if args.command == "start":
            self.flask = {
                "flask_app": args.flask_app,
                "flask_debug": args.flask_debug,
                "flask_port": args.flask_port,
                "flask_host": args.flask_host,
                "hide_flask_shell": args.hide_flask_shell,
            }
            self.huey = {
                "thread_count": args.thread_count,
                "hide_huey_shell": args.hide_huey_shell,
            }
            # self.logs_dir = args.logs
        self.processes = []

    def print_config(self):
        """Pring configuration settings for the Ripple Manager."""
        print("Flask API Configuration:", self.flask)
        print("Huey Consumer Configuration:", self.huey)
        print("Logs Directory:", os.getcwd())

    def _handle_remove_readonly(self, func, path, exc_info):
        """Handle the read-only file deletion error."""
        import stat

        os.chmod(path, stat.S_IWRITE)
        func(path)

    def start(self):
        """Start the Ripple API and Huey consumer in separate terminals."""
        if sys.platform != "win32":
            raise SystemError("API can only be run from a windows system")

        huey_consumer_path = shutil.which("huey_consumer.py")

        if not huey_consumer_path:
            print("Error: huey consumer script was not discoverable.")
            exit(1)

        # if os.path.exists(self.logs_dir):
        #     shutil.rmtree(self.logs_dir, onerror=self._handle_remove_readonly)

        # if os.path.exists(self.logs_dir):
        #     print(f"Error: could not delete {self.logs_dir}")
        #     exit(1)

        # os.makedirs(self.logs_dir, exist_ok=True)
        # if not os.path.exists(self.logs_dir):
        #     print(f"Error: could not create {self.logs_dir}")
        #     exit(1)

        python_executable = sys.executable

        print("Starting ripple1d-huey")
        huey_command = [
            python_executable,
            "-u",
            huey_consumer_path,
            "ripple1d.api.tasks.huey",
            "-w",
            str(self.huey["thread_count"]),
        ]
        if self.huey["hide_huey_shell"]:
            subprocess.Popen(["start", "cmd", "/k", " ".join(huey_command)], shell=False)

        else:
            subprocess.Popen(["start", "cmd", "/k", " ".join(huey_command)], shell=True)

        os.environ["FLASK_APP"] = self.flask["flask_app"]

        print("Starting ripple1d-flask")
        flask_command = [
            python_executable,
            "-m",
            "flask",
            "run",
            "--host",
            self.flask["flask_host"],
            "--port",
            str(self.flask["flask_port"]),
        ]

        if self.flask["flask_debug"]:
            print("Flask debug mode enabled")
            flask_command.append("--debug")

        if self.flask["hide_flask_shell"]:
            flask_result = subprocess.Popen(["start", "cmd", "/k", " ".join(flask_command)], shell=False)
        else:
            flask_result = subprocess.Popen(["start", "cmd", "/k", " ".join(flask_command)], shell=True)

        # Check if Flask started successfully
        try:
            flask_result.wait(timeout=5)
            if flask_result.returncode != 0:
                raise subprocess.CalledProcessError(flask_result.returncode, flask_command, stderr=flask_result.stderr)
            self.processes.append((flask_result, "flask"))
            print("Success!")
        except subprocess.TimeoutExpired:
            self.processes.append((flask_result, "flask"))
        except subprocess.CalledProcessError as e:
            error_message = e.stderr.read().decode() if e.stderr else "No error message captured."
            print(f"Error: Failed to start Flask. Error message: {error_message}")
            self.processes.append((None, "flask"))


def main():
    """Run Ripple in API mode."""
    parser = argparse.ArgumentParser(description="Ripple Manager Configuration")
    subparsers = parser.add_subparsers(dest="command", help="Subcommands")

    # Define the 'start' subcommand
    start_parser = subparsers.add_parser("start", help="Start the Ripple Manager")
    start_parser.add_argument(
        "--hide_flask_shell", action="store_true", help="Launch terminal for Flask API (default: False)"
    )

    start_parser.add_argument("--flask_debug", action="store_true", help="Debug mode for Flask API (default: False)")
    start_parser.add_argument("--flask_port", type=int, default=80, help="Port for Flask API (default: 80)")
    start_parser.add_argument("--flask_host", type=str, default="0.0.0.0", help="Host for Flask API (default: 0.0.0.0)")
    start_parser.add_argument(
        "--flask_app", type=str, default="ripple1d.api.app", help="Flask App (default: ripple1d.api.app)"
    )
    start_parser.add_argument("--thread_count", type=int, default=1, help="Thread count for Huey Consumer (default: 1)")
    start_parser.add_argument(
        "--hide_huey_shell", action="store_true", help="Launch terminal for Huey Consumer (default: False)"
    )
    # start_parser.add_argument(
    #     "--logs",
    #     type=str,
    #     default=os.path.join(os.getcwd(), "logs"),
    #     help="Logs directory (default: current directory/logs)",
    # )

    args = parser.parse_args()

    if args.command == "start":
        manager = RippleManager(args)
        manager.print_config()
        manager.start()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



"""huey instance for huey + Flask REST API (to be called by huey_consumer.py)."""

import json
import os
import time
import typing

import huey.signals as sigs
from huey import SqliteHuey
from huey.api import Result

from ripple1d.api.log import initialize_log
from ripple1d.api.utils import tracerbacker

LOG = initialize_log()

huey = SqliteHuey(filename=os.path.join("huey.db"), results=True)

# Create custom table task_status
huey.storage.sql(
    """
    create table if not exists "task_status" (
        "task_id" text not null,
        "func_name" text not null,
        "func_kwargs" text not null,
        "huey_status" text not null,
        "ogc_status" text not null,
        "accept_time" timestamptz not null,
        "dismiss_time" timestamptz,
        "start_time" timestamptz,
        "finish_time" timestamptz,
        "status_time" timestamptz,
        "finish_duration_minutes" real,
        primary key("task_id")
    );
    """,
    commit=True,
)

# Create custom triggers to keep status_time field updated
huey.storage.sql(
    """
    create trigger if not exists "trg_task_status_insert" after insert on "task_status"
    begin
        update "task_status" set
            "status_time" = max(
            ifnull("accept_time", datetime(0)),
            ifnull("dismiss_time", datetime(0)),
            ifnull("start_time", datetime(0)),
            ifnull("finish_time", datetime(0))
        ),
            "finish_duration_minutes" = 1440.0 * (julianday("finish_time") - julianday("start_time"))
        where "task_id" = NEW."task_id";
    end;
    """,
    commit=True,
)
huey.storage.sql(
    """
    create trigger if not exists "trg_task_status_update" after update on "task_status"
    begin
        update "task_status" set
            "status_time" = max(
            ifnull("accept_time", datetime(0)),
            ifnull("dismiss_time", datetime(0)),
            ifnull("start_time", datetime(0)),
            ifnull("finish_time", datetime(0))
        ),
            "finish_duration_minutes" = 1440.0 * (julianday("finish_time") - julianday("start_time"))
        where "task_id" = NEW."task_id";
    end;
    """,
    commit=True,
)


def create_and_enqueue_task(func: typing.Callable, kwargs: dict = {}) -> Result:
    """Create a task instance, then add it to the custom table "task_status" before adding it to the queue.

    Return the Result object associated with the task, which is the same class type that would be returned by
    calling the @huey.task()-decorated function itself (in this case, `_process`).

    This is needed to avoid a race condition, this ensures that the task exists in the custom table before
    it starts executing.
    """
    task_instance = _process.s(func, kwargs)
    huey.storage.sql(
        """
    insert into "task_status"
        ("task_id", "func_name", "func_kwargs", "huey_status", "ogc_status", "accept_time")
    values
        (?, ?, ?, 'queued', 'accepted', datetime('now'))""",
        (task_instance.id, func.__name__, json.dumps(kwargs)),
        True,
    )
    return huey.enqueue(task_instance)


def revoke_task(task_id: str):
    """Revoke a task."""
    huey.revoke_by_id(task_id)
    expression = f"""
    update "task_status"
    set
        "ogc_status" = 'dismissed',
        "dismiss_time" = datetime('now')
    where "task_id" = ?
    """
    args = (task_id,)
    huey.storage.sql(expression, args, True)


def task_status(only_task_id: str | None) -> dict[str, dict]:
    """Return dictionary of tasks, where key is task ID and value is subdict (fields related to each task).

    If only_task_id is None, all tasks will be returned.
    If only_task_id is *not* None, only the provided task will be returned.
    """
    expression = """
        select
            "task_id",
            "func_name",
            "func_kwargs",
            "huey_status",
            "ogc_status",
            "accept_time",
            "dismiss_time",
            "start_time",
            "finish_time",
            "status_time",
            "finish_duration_minutes"
        from "task_status"
        order by "status_time" desc
        """
    results_raw = huey.storage.sql(expression, results=True)
    results_dict = {}
    for (
        task_id,
        func_name,
        func_kwargs,
        huey_status,
        ogc_status,
        accept_time,
        dismiss_time,
        start_time,
        finish_time,
        status_time,
        finish_duration_minutes,
    ) in results_raw:
        if only_task_id is not None and task_id != only_task_id:
            continue

        results_dict[task_id] = {
            "huey_status": huey_status,
            "func_name": func_name,
            "func_kwargs": func_kwargs,
            "ogc_status": ogc_status,
            "accept_time": accept_time,
            "dismiss_time": dismiss_time,
            "start_time": start_time,
            "finish_time": finish_time,
            "status_time": status_time,
            "finish_duration_minutes": finish_duration_minutes,
        }
    return results_dict


def huey_status(task_id: str) -> str:
    """For given task ID, return its current status in huey terms."""
    expression = """select "huey_status" from "task_status" where "task_id" = ?"""
    results = huey.storage.sql(expression, (task_id,), results=True)
    if not results:
        return "notfound"
    return results[0][0]


def ogc_status(task_id: str) -> str:
    """For given task ID, return its current status in OGC API terms.

    WARNING assumes function executed by task is wrapped by `tracerbacker`
    """
    expression = """select "ogc_status" from "task_status" where "task_id" = ?"""
    results = huey.storage.sql(expression, (task_id,), results=True)
    if not results:
        return "notfound"
    return results[0][0]


def noop():
    """Do nothing except log a message. For ping endpoint and testing."""
    LOG.info("This message is from the noop function")
    pass


def sleep15():
    """Do nothing except log a message and then sleep for a while.  For testing."""
    LOG.info("This message is from the sleep15 function")
    time.sleep(15)
    return ("Slept for 15", 123.456)


# If needing the worker to know about its own task, then use `@huey.task(context=True)``
# and add `task=None` to the task function definition.
@huey.task()
@tracerbacker
def _process(func: typing.Callable, kwargs: dict = {}):
    """Execute generic huey task that calls the provided func with provided kwargs, asynchronously."""
    return func(**kwargs)


@huey.signal()
def _handle_signals(signal, task, exc=None):
    """Update the status in the task_status table When task emits a signal."""
    match signal:
        case sigs.SIGNAL_EXECUTING:
            time_field = "start_time"
            ogc_status = "running"

        case sigs.SIGNAL_COMPLETE | sigs.SIGNAL_ERROR:
            time_field = "finish_time"
            tracerbacker_return = huey.result(task.id, preserve=True)
            if tracerbacker_return is None:
                ogc_status = "notfound"
            else:
                if tracerbacker_return["err"] is None:
                    ogc_status = "successful"
                else:
                    ogc_status = "failed"

        case sigs.SIGNAL_CANCELED | sigs.SIGNAL_LOCKED | sigs.SIGNAL_EXPIRED | sigs.SIGNAL_INTERRUPTED:
            time_field = "dismiss_time"
            ogc_status = "dismissed"

        case sigs.SIGNAL_REVOKED:
            # Set huey status, then short-circuit without setting "ogc_status" or time field.
            # OGC status and dismiss time field are handled at time of revoke call, rather than waiting
            # for the revoke signal which happens later.
            huey.storage.sql(
                """update "task_status" set huey_status = ? where "task_id" = ?""", (signal, task.id), True
            )
            return

        case _:  # e.g. SIGNAL_RETRYING, SIGNAL_SCHEDULED
            raise ValueError(f"Unhandled signal: {signal}")

    expression = f"""
        update "task_status"
        set
            "huey_status" = ?,
            "ogc_status" = ?,
            "{time_field}" = datetime('now')
        where "task_id" = ?
        """

    args = (signal, ogc_status, task.id)
    huey.storage.sql(expression, args, True)



"""API Lofging module for Ripple1D."""

import inspect
import json
import logging
import os
import traceback
from datetime import datetime, timezone

from ripple1d.ripple1d_logger import RippleLogFormatter

LOGS: dict[str, logging.Logger] = {}  # global that is modified by initialize_log()


def initialize_log(log_dir: str = "") -> logging.Logger:
    """Initialize log with JSON-LD style formatting and throttled level for AWS libs.

    By default sends to StreamHandler (stdout/stderr), but can provide a filename to log to disk instead.
    """
    global LOGS

    filename = os.path.join(log_dir, f"{_get_log_filename_prefix()}-{_get_log_filename_suffix()}.jsonld")

    # If this log has already been initialized, just return it
    if filename in LOGS:
        return LOGS[filename]

    logging.getLogger("boto3").setLevel(logging.WARNING)
    logging.getLogger("botocore").setLevel(logging.WARNING)

    log = logging.getLogger()
    log.setLevel(logging.INFO)
    formatter = RippleLogFormatter()

    if filename:
        print(f"Initializing log file: {filename}")
        os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
        file_handler = logging.FileHandler(filename=filename)
        file_handler.setFormatter(formatter)
        log.addHandler(file_handler)
    else:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        log.addHandler(stream_handler)

    # Check again if the log has already been initialized, this time just before assigning the new log, to minimize
    # chance of race condition.
    if filename in LOGS:
        return LOGS[filename]

    LOGS[filename] = log
    return log


def _get_log_filename_prefix():
    """Return a string of the current UTC timestamp, to be used as the left-hand portion of the log file name."""
    return f"{datetime.now(tz=timezone.utc).isoformat().replace(':','-')}-ripple1d"


def _get_log_filename_suffix():
    stack_filenames = [frame.filename for frame in inspect.stack()]

    # Check for specific patterns in the stack filenames
    if any("huey" in filename for filename in stack_filenames):
        return "huey"
    elif any("flask" in filename for filename in stack_filenames):
        return "flask"
    else:
        raise ValueError(
            f"Could not determine if process invoked by huey or by flask. Stack filenames: {stack_filenames}"
        )



"""Run API tests."""

import json
from http import HTTPStatus

import requests
from dotenv import dotenv_values

FLASK_ENV_VARS = dotenv_values(".flaskenv")


def ping():
    """
    Send a GET request to the /ping endpoint of a local server.

    This function constructs the URL for the /ping endpoint using the FLASK_RUN_PORT
    specified in the .flaskenv file. It then sends a GET request to this URL.

    If the response status code is not HTTPStatus.OK (200), it raises an HTTPError with
    the response content and status code. Otherwise, it prints the JSON-decoded response content.
    """
    url = f"http://localhost:{FLASK_ENV_VARS['FLASK_RUN_PORT']}/ping"
    print(f"Pinging at url {repr(url)}")
    r = requests.get(url=url)
    if r.status_code != HTTPStatus.OK:
        raise requests.HTTPError(r.content.decode(), r.status_code)
    response_data = json.loads(r.content)
    print(response_data)


def execute_test_process():
    """
    Send a POST request to initiate a test process on a local server.

    This function constructs the URL for initiating a test process using the FLASK_RUN_PORT
    specified in the .flaskenv file. It then sends a POST request with an empty JSON payload
    to this URL.

    If the response status code is not HTTPStatus.OK (200), it raises an HTTPError with
    the response content and status code. Otherwise, it prints the JSON-decoded response content.
    """
    url = f"http://localhost:{FLASK_ENV_VARS['FLASK_RUN_PORT']}/processes/test/execution"
    print(f"Executing test process at url {repr(url)}")
    r = requests.post(url=url, json={})
    if r.status_code != HTTPStatus.OK:
        raise requests.HTTPError(r.content.decode(), r.status_code)
    response_data = json.loads(r.content)
    print(response_data)


def main():
    """
    Execute the ping and test process functions.

    This function first calls the ping() function to send a GET request to the /ping endpoint.
    Then, it calls the execute_test_process() function to initiate a test process by sending
    a POST request to the relevant endpoint.
    """
    ping()
    execute_test_process()


if __name__ == "__main__":
    main()



"""Extract geospatial data from HEC-RAS files."""

import glob
import json
import logging
import os
import shutil
import sqlite3
import tempfile
from pathlib import Path, PurePosixPath

import boto3
import geopandas as gpd
import pandas as pd
import pystac
from botocore.exceptions import ParamValidationError
from pyproj import CRS

import ripple1d
from ripple1d.data_model import NwmReachModel, RippleSourceModel
from ripple1d.errors import CouldNotIdentifyPrimaryPlanError, NoFlowFileSpecifiedError
from ripple1d.ras import VALID_GEOMS, VALID_STEADY_FLOWS, RasFlowText, RasGeomText, RasManager, RasPlanText, RasProject
from ripple1d.utils.dg_utils import bbox_to_polygon
from ripple1d.utils.gpkg_utils import (
    create_geom_item,
    create_thumbnail_from_gpkg,
    get_asset_info,
    get_river_miles,
    gpkg_to_geodataframe,
    reproject,
    write_thumbnail_to_s3,
)
from ripple1d.utils.ripple_utils import get_path, prj_is_ras, xs_concave_hull
from ripple1d.utils.s3_utils import (
    get_basic_object_metadata,
    init_s3_resources,
    list_keys,
    s3_key_public_url_converter,
    str_from_s3,
)
from ripple1d.utils.sqlite_utils import create_non_spatial_table


def geom_flow_to_gpkg(
    ras_project: RasProject, crs: str, gpkg_file: str, metadata: dict, client: boto3.client = None, bucket: str = None
) -> None:
    """Write geometry and flow data to a geopackage."""
    layers, metadata = geom_flow_to_gdfs(ras_project, crs, metadata, client, bucket)
    for layer, gdf in layers.items():
        if layer == "XS":
            if "Junction" in layers.keys():
                xs_concave_hull(gdf, layers["Junction"]).to_file(gpkg_file, driver="GPKG", layer="XS_concave_hull")
            else:
                xs_concave_hull(gdf).to_file(gpkg_file, driver="GPKG", layer="XS_concave_hull")
        gdf.to_file(gpkg_file, driver="GPKG", layer=layer)
    create_non_spatial_table(gpkg_file, metadata)
    return metadata


def find_a_valid_file(
    directory: str, valid_extensions: list[str], client: boto3.client = None, bucket: str = None
) -> str:
    """Find a file in the directory that contains a valid extension. Returns the first valid file found."""
    if client and bucket:
        paths = list_keys(client, bucket, directory)
    else:
        paths = glob.glob(f"{directory}/*")
    for path in paths:
        if Path(path).suffix in valid_extensions:
            return path


def gather_metdata(metadata: dict, ras_project: RasProject, rp: RasPlanText, rf: RasFlowText, rg: RasGeomText) -> dict:
    """Gather metadata from the ras project and its components."""
    metadata["plans_files"] = "\n".join(
        [get_path(i).replace(f"{ras_project._ras_dir}\\", "") for i in ras_project.plans if get_path(i)]
    )
    metadata["geom_files"] = "\n".join(
        [get_path(i).replace(f"{ras_project._ras_dir}\\", "") for i in ras_project.geoms if get_path(i)]
    )
    metadata["steady_flow_files"] = "\n".join(
        [get_path(i).replace(f"{ras_project._ras_dir}\\", "") for i in ras_project.steady_flows if get_path(i)]
    )
    metadata["unsteady_flow_files"] = "\n".join(
        [get_path(i).replace(f"{ras_project._ras_dir}\\", "") for i in ras_project.unsteady_flows if get_path(i)]
    )
    metadata["ras_project_file"] = ras_project._ras_text_file_path.replace(f"{ras_project._ras_dir}\\", "")
    metadata["ras_project_title"] = ras_project.title
    metadata["plans_titles"] = "\n".join(
        [RasPlanText(get_path(i), rg.crs).title for i in ras_project.plans if get_path(i)]
    )
    metadata["geom_titles"] = "\n".join(
        [RasGeomText(get_path(i), rg.crs).title for i in ras_project.geoms if get_path(i)]
    )
    metadata["steady_flow_titles"] = "\n".join(
        [RasFlowText(get_path(i)).title for i in ras_project.steady_flows if get_path(i)]
    )
    metadata["active_plan"] = get_path(ras_project._ras_root_path + ras_project.current_plan).replace(
        f"{ras_project._ras_dir}\\", ""
    )
    metadata["primary_plan_file"] = get_path(rp._ras_text_file_path).replace(f"{ras_project._ras_dir}\\", "")
    metadata["primary_plan_title"] = rp.title
    metadata["primary_flow_file"] = get_path(rf._ras_text_file_path).replace(f"{ras_project._ras_dir}\\", "")
    metadata["primary_geom_file"] = get_path(rg._ras_text_file_path).replace(f"{ras_project._ras_dir}\\", "")
    metadata["primary_geom_title"] = rg.title
    metadata["primary_flow_title"] = rf.title
    if len(rg.version) >= 1:
        metadata["ras_version"] = rg.version[0]
    else:
        metadata["ras_version"] = None
    metadata["ripple1d_version"] = ripple1d.__version__
    fcls = pd.DataFrame(rf.flow_change_locations)
    metadata["profile_names"] = "\n".join(fcls["profile_names"].iloc[0])
    metadata["units"] = ras_project.units
    return metadata


def geom_flow_to_gdfs(
    ras_project: RasProject, crs: str, metadata: dict, client: boto3.client = None, bucket: str = None
) -> tuple:
    """Write geometry and flow data to a geopackage."""
    if client and bucket:
        rp = detemine_primary_plan(ras_project, crs, ras_project._ras_text_file_path, client, bucket)
        # get steady flow file
        try:
            plan_steady_file = get_path(rp.plan_steady_file, client, bucket)
        except NoFlowFileSpecifiedError as e:
            logging.warning(e)
            plan_steady_file = find_a_valid_file(ras_project._ras_dir, VALID_STEADY_FLOWS, client, bucket)

        # get geometry file
        try:
            plan_geom_file = get_path(rp.plan_geom_file, client, bucket)
        except NoFlowFileSpecifiedError:
            logging.warning(e)
            plan_geom_file = find_a_valid_file(ras_project._ras_dir, VALID_GEOMS, client, bucket)

        string = str_from_s3(plan_steady_file, client, bucket)
        rf = RasFlowText.from_str(string, " .f01")

        string = str_from_s3(plan_geom_file, client, bucket)
        rg = RasGeomText.from_str(string, crs, " .g01")

    else:
        rp = detemine_primary_plan(ras_project, crs, ras_project._ras_text_file_path)

        try:
            plan_steady_file = get_path(rp.plan_steady_file)
        except NoFlowFileSpecifiedError as e:
            logging.warning(e)
            plan_steady_file = find_a_valid_file(ras_project._ras_dir, VALID_STEADY_FLOWS)

        try:
            plan_geom_file = get_path(rp.plan_geom_file)
        except NoFlowFileSpecifiedError as e:
            logging.warning(e)
            plan_geom_file = find_a_valid_file(ras_project._ras_dir, VALID_GEOMS)

        rf = RasFlowText(plan_steady_file)
        rg = RasGeomText(plan_geom_file, crs)

    layers = {}
    if rg.cross_sections:
        xs_gdf = rg.xs_gdf
        if "u" in Path(plan_steady_file).suffix:
            xs_gdf["flow_tile"] = rf.title
        else:
            xs_gdf = geom_flow_xs_gdf(rg, rf, xs_gdf)
        xs_gdf["plan_title"] = rp.title
        xs_gdf["geom_title"] = rg.title
        if len(rg.version) >= 1:
            xs_gdf["version"] = rg.version[0]
        else:
            xs_gdf["version"] = None
        xs_gdf["units"] = ras_project.units
        xs_gdf["project_title"] = ras_project.title
        layers["XS"] = xs_gdf

    if rg.reaches:
        layers["River"] = rg.reach_gdf

    if rg.junctions:
        layers["Junction"] = rg.junction_gdf

    if rg.structures:
        layers["Structure"] = rg.structures_gdf

    return layers, gather_metdata(metadata, ras_project, rp, rf, rg)


def geom_flow_xs_gdf(rg: RasGeomText, rf: RasFlowText, xs_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Create a geodataframe with cross section geometry and flow data."""
    xs_gdf[["flows", "profile_names"]] = None, None

    fcls = pd.DataFrame(rf.flow_change_locations)
    fcls["river_reach"] = fcls["river"] + fcls["reach"]

    for river_reach in fcls["river_reach"].unique():
        # get flow change locations for this reach
        fcls_rr = fcls.loc[fcls["river_reach"] == river_reach, :].sort_values(by="rs", ascending=False)

        # iterate through this reaches flow change locations and set cross section flows/profile names
        for _, row in fcls_rr.iterrows():
            # add flows to xs_gdf
            xs_gdf.loc[
                (xs_gdf["river"] == row["river"])
                & (xs_gdf["reach"] == row["reach"])
                & (xs_gdf["river_station"] <= row["rs"]),
                "flows",
            ] = "\n".join([str(f) for f in row["flows"]])

            # add profile names to xs_gdf
            xs_gdf.loc[
                (xs_gdf["river"] == row["river"])
                & (xs_gdf["reach"] == row["reach"])
                & (xs_gdf["river_station"] <= row["rs"]),
                "profile_names",
            ] = "\n".join(row["profile_names"])

    xs_gdf["flow_title"] = rf.title
    return xs_gdf


def detemine_primary_plan(
    ras_project: RasProject,
    crs: str,
    ras_text_file_path: str,
    client: boto3.session.Session.client = None,
    bucket: str = None,
) -> RasPlanText:
    """
    Determine the primary plan for a ras project.

    Exammple: The active plan if it does not contain encroachments.
    If the active plan contains encroachments, the first plan without encroachments is returned.
    If no plans are found without encroachments, an error is raised.
    """
    if len(ras_project.plans) == 1:
        if client:
            plan_path = get_path(ras_project.plans[0], client, bucket)
            string = str_from_s3(plan_path, client, bucket)
            return RasPlanText.from_str(string, crs, plan_path)
        else:
            plan_path = get_path(ras_project.plans[0])
            return RasPlanText(plan_path, crs)
    candidate_plans = []
    for plan_path in ras_project.plans:
        plan_path = get_path(ras_project.plans[0], client, bucket)
        if client:
            try:
                string = str_from_s3(plan_path, client, bucket)
            except ParamValidationError as e:
                logging.warning(f"Missing plan file {plan_path} in {ras_project._ras_text_file_path}")
                continue
            if not string.__contains__("Encroach Node"):
                candidate_plans.append(RasPlanText.from_str(string, crs, plan_path))
        else:
            if os.path.exists(plan_path):
                with open(plan_path) as src:
                    string = src.read()
                if not string.__contains__("Encroach Node"):
                    candidate_plans.append(RasPlanText.from_str(string, crs, plan_path))
    if len(candidate_plans) > 1 or not candidate_plans:
        plan_path = ras_project._ras_root_path + "." + ras_project.current_plan.lstrip(".")
        if client:
            string = str_from_s3(plan_path, client, bucket)
            return RasPlanText.from_str(string, crs, plan_path)
        else:
            return RasPlanText(plan_path, crs)
        # raise CouldNotIdentifyPrimaryPlanError(f"Could not identfiy a primary plan for {ras_text_file_path}")
    else:
        return candidate_plans[0]


def gpkg_from_ras(source_model_directory: str, crs: str, metadata: dict):
    """Write geometry and flow data to a geopackage locally."""
    prjs = glob.glob(f"{source_model_directory}/*.prj")
    ras_text_file_path = None

    for prj in prjs:
        if prj_is_ras(prj):
            ras_text_file_path = prj
            break

    if not ras_text_file_path:
        raise FileNotFoundError(f"No ras project file found in {source_model_directory}")

    output_gpkg_path = ras_text_file_path.replace(".prj", ".gpkg")
    rp = RasProject(ras_text_file_path)
    return geom_flow_to_gpkg(rp, crs, output_gpkg_path, metadata)


def gpkg_from_ras_s3(s3_prefix: str, crs: str, metadata: dict, bucket: str):
    """Write geometry and flow data to a geopackage on s3."""
    _, client, _ = init_s3_resources()

    # make temp directory
    temp_dir = tempfile.mkdtemp()
    temp_path = os.path.join(temp_dir, "temp.gpkg")

    ras_text_file_path = None
    for key in list_keys(client, bucket, s3_prefix, ".prj"):
        string = str_from_s3(key, client, bucket)
        if "Proj Title" in string.split("\n")[0]:
            ras_text_file_path = key
            break

    if not ras_text_file_path:
        raise FileNotFoundError(f"No ras project file found in {s3_prefix}")

    # read ras project file get list of plans
    ras_project = RasProject.from_str(string, ras_text_file_path)

    geom_flow_to_gpkg(ras_project, crs, temp_path, metadata, client, bucket)

    # move geopackage to s3
    output_gpkg_path = ras_text_file_path.replace(f"s3://{bucket}/", "").replace(".prj", ".gpkg")
    logging.debug(f"uploading {output_gpkg_path} to s3")
    client.upload_file(
        Bucket=bucket,
        Key=output_gpkg_path,
        Filename=temp_path,
    )
    shutil.rmtree(temp_dir)


def new_stac_item_s3(
    gpkg_s3_key: str,
    new_stac_item_s3_key: str,
    thumbnail_png_s3_key: str,
    s3_ras_project_key: str,
    bucket: str,
    mip_case_no: str = None,
    dev_mode: bool = False,
):
    """Create a new stac item from a geopackage on s3."""
    logging.debug("Creating item from gpkg")
    # Instantitate S3 resources

    if mip_case_no is None:
        mip_case_no = "N/A"

    session, s3_client, s3_resource = init_s3_resources()
    item_basename = Path(s3_ras_project_key).name
    item_id = item_basename.replace(".prj", "")
    prefix = Path(s3_ras_project_key).parent.as_posix()
    asset_list = list_keys(s3_client, bucket, prefix)

    gdfs = gpkg_to_geodataframe(f"s3://{bucket}/{gpkg_s3_key}")
    river_miles = get_river_miles(gdfs["River"])
    crs = gdfs["River"].crs
    gdfs = reproject(gdfs)

    logging.debug("Creating png thumbnail")
    fig = create_thumbnail_from_gpkg(gdfs)
    write_thumbnail_to_s3(fig, thumbnail_png_s3_key, bucket, s3_client)

    # Create item
    bbox = pd.concat(gdfs).total_bounds
    footprint = bbox_to_polygon(bbox)

    data = gdfs["XS"].iloc[0]
    properties = {
        "ripple: version": ripple1d.__version__,
        "ras version": data["version"],
        "project title": data["project_title"],
        "plan title": data["plan_title"],
        "geom title": data["geom_title"],
        "flow title": data["flow_title"],
        # "profile names": data["profile_names"].splitlines(),
        "MIP:case_ID": mip_case_no,
        "river miles": str(river_miles),
        "proj:wkt2": crs.to_wkt(),
        "proj:epsg": crs.to_epsg(),
    }

    item = create_geom_item(item_id, bbox, footprint, properties)
    rsm = RippleSourceModel(s3_ras_project_key, crs)

    asset_list = asset_list + [thumbnail_png_s3_key, gpkg_s3_key]
    for asset_key in asset_list:
        obj = s3_resource.Bucket(bucket).Object(asset_key)
        metadata = get_basic_object_metadata(obj)

        asset_info = get_asset_info(asset_key, rsm, bucket)
        if asset_key == thumbnail_png_s3_key:
            asset = pystac.Asset(
                s3_key_public_url_converter(f"s3://{bucket}/{asset_key}"),
                extra_fields=metadata,
                roles=asset_info["roles"],
                description=asset_info["description"],
            )
        else:
            asset = pystac.Asset(
                f"s3://{bucket}/{asset_key}",
                extra_fields=metadata,
                roles=asset_info["roles"],
                description=asset_info["description"],
            )
        item.add_asset(asset_info["title"], asset)

    s3_client.put_object(
        Body=json.dumps(item.to_dict()).encode(),
        Bucket=bucket,
        Key=new_stac_item_s3_key,
    )

    logging.debug("Program completed successfully")


def new_stac_item(ras_project_directory: str, ras_s3_prefix: str):
    """Create a new stac item from a geopackage locally ."""
    logging.debug("Creating item from gpkg")

    nwm_rm = NwmReachModel(ras_project_directory)
    rm = RasManager(nwm_rm.ras_project_file, crs=nwm_rm.crs)
    gdfs = gpkg_to_geodataframe(nwm_rm.ras_gpkg_file)

    river_miles = get_river_miles(gdfs["River"])
    crs = gdfs["River"].crs
    gdfs = reproject(gdfs)

    logging.debug("Creating png thumbnail")
    fig = create_thumbnail_from_gpkg(gdfs)
    fig.savefig(nwm_rm.thumbnail_png)

    # Create item
    bbox = pd.concat(gdfs).total_bounds
    footprint = bbox_to_polygon(bbox)

    data = gdfs["XS"].iloc[0]
    properties = {
        "ripple: version": ripple1d.__version__,
        "ras version": rm.version,
        "ras_units": rm.ras_project.units,
        "project title": rm.ras_project.title,
        "plan titles": {key: val.file_extension for key, val in rm.plans.items()},
        "geom titles": {key: val.file_extension for key, val in rm.geoms.items()},
        "flow titles": {key: val.file_extension for key, val in rm.flows.items()},
        "river miles": str(river_miles),
        "NWM to_id": nwm_rm.ripple1d_parameters["nwm_to_id"],
        "proj:wkt2": crs.to_wkt(),
        "proj:epsg": crs.to_epsg(),
    }
    item = create_geom_item(nwm_rm.model_name, bbox, footprint, properties)

    for asset_key in nwm_rm.assets:

        asset_info = get_asset_info(asset_key, nwm_rm)
        if ras_s3_prefix:
            asset_key = str(PurePosixPath(Path(asset_key.replace(nwm_rm.model_directory, ras_s3_prefix))))
        asset = pystac.Asset(
            os.path.relpath(asset_key),
            extra_fields=asset_info["extra_fields"],
            roles=asset_info["roles"],
            description=asset_info["description"],
        )
        item.add_asset(asset_info["title"], asset)
    item.add_derived_from(nwm_rm.ripple1d_parameters["source_model"])
    item.add_derived_from(nwm_rm.ripple1d_parameters["source_terrain"])
    item.add_derived_from(nwm_rm.ripple1d_parameters["source_nwm_reach"])
    with open(nwm_rm.model_stac_json_file, "w") as dst:
        dst.write(json.dumps(item.to_dict()))

    logging.debug("Program completed successfully")



"""Subset a geopackage based on clonfation with NWM hydrofabric."""

import json
import logging
import os

import fiona
import geopandas as gpd
import pandas as pd
from shapely.geometry import LineString

import ripple1d
from ripple1d.consts import METERS_PER_FOOT
from ripple1d.data_model import NwmReachModel, RippleSourceDirectory, RippleSourceModel
from ripple1d.utils.ripple_utils import xs_concave_hull


class RippleGeopackageSubsetter:
    """Subset a geopackage based on conflation with NWM hydrofabric."""

    def __init__(self, src_gpkg_path: str, conflation_json: str, dst_project_dir: str, nwm_id: str):

        self.src_gpkg_path = src_gpkg_path
        self.conflation_json = conflation_json
        self.dst_project_dir = dst_project_dir
        self.nwm_id = nwm_id

    @property
    def conflation_parameters(self) -> dict:
        """Extract conflation parameters from the conflation json."""
        with open(self.conflation_json, "r") as f:
            return json.load(f)

    @property
    def ripple1d_parameters(self) -> dict:
        """Extract ripple1d parameters from the conflation json."""
        return self.conflation_parameters["reaches"][self.nwm_id]

    @property
    def us_reach(self) -> str:
        """Extract upstream reach from conflation parameters."""
        return self.ripple1d_parameters["us_xs"]["reach"]

    @property
    def us_river(self) -> str:
        """Extract upstream river from conflation parameters."""
        return self.ripple1d_parameters["us_xs"]["river"]

    @property
    def us_rs(self) -> str:
        """Extract upstream river station from conflation parameters."""
        return self.ripple1d_parameters["us_xs"]["xs_id"]

    @property
    def ds_river(self) -> str:
        """Extract downstream river from conflation parameters."""
        return self.ripple1d_parameters["ds_xs"]["river"]

    @property
    def ds_reach(self) -> str:
        """Extract downstream reach from conflation parameters."""
        return self.ripple1d_parameters["ds_xs"]["reach"]

    @property
    def ds_rs(self) -> str:
        """Extract downstream river station from conflation parameters."""
        return self.ripple1d_parameters["ds_xs"]["xs_id"]

    @property
    def source_xs(self) -> gpd.GeoDataFrame:
        """Extract cross sections from the source geopackage."""
        xs = gpd.read_file(self.src_gpkg_path, layer="XS")
        return xs[xs.intersects(self.source_river.union_all())]

    @property
    def source_river(self) -> gpd.GeoDataFrame:
        """Extract river geometry from the source geopackage."""
        return gpd.read_file(self.src_gpkg_path, layer="River")

    @property
    def source_structure(self) -> gpd.GeoDataFrame:
        """Extract structures from the source geopackage."""
        if "Structure" in fiona.listlayers(self.src_gpkg_path):
            structures = gpd.read_file(self.src_gpkg_path, layer="Structure")
            return structures[structures.intersects(self.source_river.union_all())]

    @property
    def source_junction(self) -> gpd.GeoDataFrame:
        """Extract junctions from the source geopackage."""
        if "Junction" in fiona.listlayers(self.src_gpkg_path):
            return gpd.read_file(self.src_gpkg_path, layer="Junction")
        return None

    @property
    def ripple_xs(self) -> gpd.GeoDataFrame:
        """Subset cross sections based on NWM reach."""
        return self.subset_gdfs["XS"]

    @property
    def ripple_river(self) -> gpd.GeoDataFrame:
        """Subset river geometry based on NWM reach."""
        return self.subset_gdfs["River"]

    @property
    def ripple_structure(self) -> gpd.GeoDataFrame:
        """Subset structures based on NWM reach."""
        return self.subset_gdfs["Structure"]

    @property
    def subset_gdfs(self) -> dict:
        """Subset the cross sections, structues, and river geometry for a given NWM reach."""
        # subset data
        if self.us_river == self.ds_river and self.us_reach == self.ds_reach:
            ripple_xs, ripple_structure, ripple_river = self.process_as_one_ras_reach()
        else:
            ripple_xs, ripple_structure, ripple_river = self.process_as_multiple_ras_reach()

        # check if only 1 cross section for nwm_reach
        if len(ripple_xs) <= 1:
            logging.warning(f"Only 1 cross section conflated to NWM reach {self.nwm_id}. Skipping this reach.")
            return None

        # update fields
        ripple_xs = self.update_fields(ripple_xs)
        ripple_river = self.update_fields(ripple_river)
        ripple_structure = self.update_fields(ripple_structure)

        # clip river to cross sections
        ripple_river = self.clip_river(ripple_xs, ripple_river)

        if ripple_structure is not None and len(ripple_structure) > 0:
            return {"XS": ripple_xs, "River": ripple_river, "Structure": ripple_structure}
        else:
            return {"XS": ripple_xs, "River": ripple_river}

    @property
    def ripple_gpkg_file(self) -> str:
        """Return the path to the new geopackage."""
        return self.nwm_reach_model.ras_gpkg_file

    @property
    def nwm_reach_model(self) -> NwmReachModel:
        """Return the new NWM reach model object."""
        return NwmReachModel(self.dst_project_dir)

    def write_ripple_gpkg(
        self,
    ) -> None:
        """Write the subsetted geopackage to the destination project directory."""
        os.makedirs(self.dst_project_dir, exist_ok=True)

        for layer, gdf in self.subset_gdfs.items():
            # remove lateral structures
            if layer == "Structure":
                if (gdf["type"] == 6).any():
                    logging.warning(
                        f"Lateral structures are not currently supported in ripple1d. The lateral structures will be dropped."
                    )
                    gdf = gdf.loc[gdf["type"] != 6, :]

            if gdf.shape[0] > 0:
                gdf.to_file(self.ripple_gpkg_file, layer=layer)
                if layer == "XS":
                    xs_concave_hull(gdf).to_file(self.ripple_gpkg_file, driver="GPKG", layer="XS_concave_hull")

    @property
    def min_flow(self) -> float:
        """Extract the min flow from the cross sections."""
        if "flows" in self.ripple_xs.columns:
            return self.ripple_xs["flows"].str.split("\n", expand=True).astype(float).min().min()
        else:
            logging.warning(f"no flows specified in source model gpkg for {self.nwm_id}")
            return 10000000000000

    @property
    def max_flow(self) -> float:
        """Extract the max flow from the cross sections."""
        if "flows" in self.ripple_xs.columns:
            return self.ripple_xs["flows"].str.split("\n", expand=True).astype(float).max().max()
        else:
            logging.warning(f"no flows specified in source model gpkg for {self.nwm_id}")
            return 0

    @property
    def crs(self):
        """Extract the CRS from the cross sections."""
        return self.source_xs.crs

    def walk_junctions(self) -> list[str]:
        """Check if junctions are present for the given river-reaches."""
        river_reaches = []
        if not self.source_junction.empty:
            while True:

                for _, row in self.source_junction.iterrows():

                    us_rivers = row.us_rivers.split(",")
                    us_reaches = row.us_reaches.split(",")

                    for river, reach in zip(us_rivers, us_reaches):
                        if row.ds_rivers == self.ds_river and row.ds_reaches == self.ds_reach:
                            return river_reaches
                        if river == self.us_river and reach == self.us_reach:
                            river_reaches.append(f"{row.ds_rivers.ljust(16)},{row.ds_reaches.ljust(16)}")
                            us_river = row.ds_rivers
                            us_reach = row.ds_reaches

    def clean_river_stations(self, ras_data: str) -> str:
        """Clean up river station data."""
        lines = ras_data.splitlines()
        data = lines[0].split(",")
        if "*" in data[1]:
            data[1] = str(float(data[1].rstrip("*"))) + "*"
            data[1] = data[1].ljust(8)
        else:
            data[1] = str(float(data[1])).ljust(8)
        lines[0] = ",".join(data)
        return "\n".join(lines) + "\n"

    def round_river_stations(self, ras_data: str) -> str:
        """Clean up river station data."""
        lines = ras_data.splitlines()
        data = lines[0].split(",")
        if "*" in data[1]:
            data[1] = str(float(round(float(data[1].rstrip("*"))))) + "*"
            data[1] = data[1].ljust(8)
        else:
            data[1] = str(float(round(float(data[1])))).ljust(8)
        lines[0] = ",".join(data)
        return "\n".join(lines) + "\n"

    def update_river_station(self, ras_data: str, river_station: str) -> str:
        """Update river station data."""
        lines = ras_data.splitlines()
        data = lines[0].split(",")
        if "*" in data[1]:
            data[1] = str(float(data[1].rstrip("*")) + river_station) + "*"
            data[1] = data[1].ljust(8)
        else:
            data[1] = str(float(data[1]) + river_station).ljust(8)
        lines[0] = ",".join(data)
        return "\n".join(lines) + "\n"

    def junction_length_to_reach_lengths(self):
        """Adjust reach lengths using junction."""
        # TODO adjust reach lengths using junction lengths
        raise NotImplementedError
        # for row in junction_gdf.iterrows():
        #     if us_river in row["us_rivers"] and us_reach in row["us_reach"]:

    def process_as_one_ras_reach(self) -> tuple:
        """Process as a single ras-river-reach."""
        xs_subset_gdf = self.source_xs.loc[
            (self.source_xs["river"] == self.us_river)
            & (self.source_xs["reach"] == self.us_reach)
            & (self.source_xs["river_station"] >= float(self.ds_rs))
            & (self.source_xs["river_station"] <= float(self.us_rs))
        ]
        if self.source_structure is not None:
            structures_subset_gdf = self.source_structure.loc[
                (self.source_structure["river"] == self.us_river)
                & (self.source_structure["reach"] == self.us_reach)
                & (self.source_structure["river_station"] >= float(self.ds_rs))
                & (self.source_structure["river_station"] <= float(self.us_rs))
            ]
        else:
            structures_subset_gdf = None
        river_subset_gdf = self.source_river.loc[
            (self.source_river["river"] == self.us_river) & (self.source_river["reach"] == self.us_reach)
        ]

        return xs_subset_gdf, structures_subset_gdf, river_subset_gdf

    @property
    def xs_us_reach(self) -> gpd.GeoDataFrame:
        """Extract cross sections for the upstream reach."""
        return self.source_xs.loc[
            (self.source_xs["river"] == self.us_river)
            & (self.source_xs["reach"] == self.us_reach)
            & (self.source_xs["river_station"] <= float(self.us_rs))
        ]

    @property
    def xs_ds_reach(self) -> gpd.GeoDataFrame:
        """Extract cross sections for the downstream reach."""
        return self.source_xs.loc[
            (self.source_xs["river"] == self.ds_river)
            & (self.source_xs["reach"] == self.ds_reach)
            & (self.source_xs["river_station"] >= float(self.ds_rs))
        ]

    @property
    def structures_us_reach(self) -> gpd.GeoDataFrame:
        """Extract structures for the upstream reach."""
        if self.source_structure is not None:
            return self.source_structure.loc[
                (self.source_structure["river"] == self.us_river)
                & (self.source_structure["reach"] == self.us_reach)
                & (self.source_structure["river_station"] <= float(self.us_rs))
            ]

    @property
    def structures_ds_reach(self) -> gpd.GeoDataFrame:
        """Extract structures for the downstream reach."""
        if self.source_structure is not None:
            return self.source_structure.loc[
                (self.source_structure["river"] == self.ds_river)
                & (self.source_structure["reach"] == self.ds_reach)
                & (self.source_structure["river_station"] >= float(self.ds_rs))
            ]

    def add_intermediate_river_reaches(self, xs_us_reach: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Add intermediate river reaches to the xs_us_reach."""
        xs_us_reach = self.xs_us_reach.copy()
        for intermediate_river_reach in intermediate_river_reaches:
            xs_intermediate_river_reach = self.source_xs.loc[self.source_xs["river_reach"] == intermediate_river_reach]
            if xs_us_reach["river_station"].min() <= xs_intermediate_river_reach["river_station"].max():
                logging.warning(
                    f"the lowest river station on the upstream reach ({xs_us_reach['river_station'].min()}) is less"
                    f" than the highest river station on the intermediate reach ({xs_intermediate_river_reach['river_station'].max()}) for nwm_id: {self.nwm_id}"
                )
                xs_us_reach["river_station"] = (
                    xs_us_reach["river_station"] + xs_intermediate_river_reach["river_station"].max()
                )
                xs_us_reach["ras_data"] = xs_us_reach["ras_data"].apply(
                    lambda ras_data: self.update_river_station(
                        ras_data, xs_intermediate_river_reach["river_station"].max()
                    )
                )
            xs_us_reach = pd.concat([xs_us_reach, xs_intermediate_river_reach])
        return xs_us_reach

    def adjust_river_stations(self, xs_us_reach, structures_us_reach) -> tuple:
        """Adjust river stations of the upstream reach if the min river station of the upstream reach is less than the max river station of the downstream reach."""
        if xs_us_reach["river_station"].min() <= self.xs_ds_reach["river_station"].max():
            logging.warning(
                f"the lowest river station on the upstream reach ({xs_us_reach['river_station'].min()}) is less"
                f" than the highest river station on the downstream reach ({self.xs_ds_reach['river_station'].max()}) for nwm_id: {self.nwm_id}"
            )
            xs_us_reach["river_station"] = xs_us_reach["river_station"] + self.xs_ds_reach["river_station"].max()
            xs_us_reach["ras_data"] = xs_us_reach["ras_data"].apply(
                lambda ras_data: self.update_river_station(ras_data, self.xs_ds_reach["river_station"].max())
            )
            if structures_us_reach is not None:
                structures_us_reach["river_station"] = (
                    structures_us_reach["river_station"] + self.xs_ds_reach["river_station"].max()
                )
                structures_us_reach["ras_data"] = structures_us_reach["ras_data"].apply(
                    lambda ras_data: self.update_river_station(ras_data, self.xs_ds_reach["river_station"].max())
                )
        return xs_us_reach, structures_us_reach

    def process_as_multiple_ras_reach(self) -> tuple:
        """Process as multiple ras-river-reach."""
        if "Junction" in fiona.listlayers(self.src_gpkg_path):
            junction_gdf = gpd.read_file(self.src_gpkg_path, layer="Junction")
            intermediate_river_reaches = self.walk_junctions()
        else:
            intermediate_river_reaches = None

        # add intermediate river reaches to the upstream reach
        if intermediate_river_reaches:
            xs_us_reach = self.add_intermediate_river_reaches()
        else:
            xs_us_reach = self.xs_us_reach.copy()

        # update river stations
        xs_us_reach, structures_us_reach = self.adjust_river_stations(xs_us_reach, self.structures_us_reach)
        # combine us and ds gdfs
        xs_subset_gdf = pd.concat([xs_us_reach, self.xs_ds_reach])
        river_subset_gdf = self.combine_reach_features(intermediate_river_reaches)

        if self.source_structure is not None:
            structures_subset_gdf = pd.concat([structures_us_reach, self.structures_ds_reach])
        else:
            structures_subset_gdf = None

        return xs_subset_gdf, structures_subset_gdf, river_subset_gdf

    def rename_river_reach(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Rename river, reach, and river_reach columns after the nwm reach id."""
        pd.options.mode.copy_on_write = True
        gdf["river"] = self.nwm_id
        gdf["reach"] = self.nwm_id
        gdf["river_reach"] = f"{self.nwm_id.ljust(16)},{self.nwm_id.ljust(16)}"

        return gdf

    def combine_reach_features(self, intermediate_river_reaches: list[str]) -> gpd.GeoDataFrame:
        """Combine reach coordinates and update river and reach names to be nwm id."""
        us_reach = self.source_river.loc[
            (self.source_river["river"] == self.us_river) & (self.source_river["reach"] == self.us_reach)
        ]
        ds_reach = self.source_river.loc[
            (self.source_river["river"] == self.ds_river) & (self.source_river["reach"] == self.ds_reach)
        ]

        # handle river reach coords
        coords = list(us_reach.iloc[0]["geometry"].coords)
        if intermediate_river_reaches:
            for intermediate_river_reach in intermediate_river_reaches:
                intermediate_river_reach_reach = self.source_river.loc[
                    self.source_river["river_reach"] == intermediate_river_reach
                ]
                coords += list(intermediate_river_reach_reach.iloc[0]["geometry"].coords)
        coords += list(ds_reach.iloc[0]["geometry"].coords)

        return gpd.GeoDataFrame(
            {"geometry": [LineString(coords)], "river": [self.nwm_id], "reach": [self.nwm_id]},
            geometry="geometry",
            crs=self.crs,
        )

    def update_fields(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Update fields for the new NWM reach."""
        if gdf is not None:
            gdf = self.rename_river_reach(gdf)

            # clean river stations
            if "river_station" in gdf.columns:
                if (gdf["river_station"].astype(str).str.len() > 8).any():
                    gdf["ras_data"] = gdf["ras_data"].apply(lambda ras_data: self.round_river_stations(ras_data))
                    gdf["river_station"] = gdf["river_station"].round().astype(float)
                else:
                    gdf["ras_data"] = gdf["ras_data"].apply(lambda ras_data: self.clean_river_stations(ras_data))
                    gdf["river_station"] = gdf["river_station"].astype(float)
        return gdf

    def clip_river(self, xs_subset_gdf: gpd.GeoDataFrame, river_subset_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Clip the river to the concave hull of the cross sections."""
        crs = xs_subset_gdf.crs

        buffer = 10
        while True:
            concave_hull = xs_concave_hull(xs_subset_gdf).to_crs(epsg=5070).buffer(buffer * METERS_PER_FOOT).to_crs(crs)
            clipped_river_subset_gdf = river_subset_gdf.clip(concave_hull)

            buffer += 10
            if len(clipped_river_subset_gdf) == 1 & xs_subset_gdf.intersects(
                clipped_river_subset_gdf["geometry"].iloc[0]
            ).all() & isinstance(clipped_river_subset_gdf["geometry"].iloc[0], LineString):
                return clipped_river_subset_gdf

            if buffer > 10000:
                raise ValueError(
                    f"buffer too large (>10000ft) for clipping river to concave hull of cross sections. Check data for NWM Reach: {self.nwm_id}."
                )
                break

    def update_ripple1d_parameters(self, rsd: RippleSourceDirectory):
        """Update ripple1d_parameters with results of subsetting."""
        ripple1d_parameters = self.ripple1d_parameters
        ripple1d_parameters["source_model"] = rsd.ras_project_file
        ripple1d_parameters["crs"] = self.crs.to_epsg()
        ripple1d_parameters["version"] = ripple1d.__version__
        ripple1d_parameters["high_flow"] = max([ripple1d_parameters["high_flow"], self.max_flow])
        ripple1d_parameters["low_flow"] = min([ripple1d_parameters["low_flow"], self.min_flow])
        if ripple1d_parameters["high_flow"] == self.max_flow:
            ripple1d_parameters["notes"] = ["high_flow computed from source model flows"]
        if ripple1d_parameters["low_flow"] == self.min_flow:
            ripple1d_parameters["notes"] = ["low_flow computed from source model flows"]
        return ripple1d_parameters

    def write_ripple1d_parameters(self, ripple1d_parameters: dict):
        """Write ripple1d parameters to json file."""
        with open(os.path.join(self.dst_project_dir, f"{self.nwm_id}.ripple1d.json"), "w") as f:
            json.dump(ripple1d_parameters, f, indent=4)


def extract_submodel(source_model_directory: str, submodel_directory: str, nwm_id: int):
    """Use ripple conflation data to create a new GPKG from an existing ras geopackage."""
    rsd = RippleSourceDirectory(source_model_directory)
    logging.info(f"Preparing to extract NWM ID {nwm_id} from {rsd.ras_project_file}")

    if not rsd.file_exists(rsd.ras_gpkg_file):
        raise FileNotFoundError(f"cannot find file ras-geometry file {rsd.ras_gpkg_file}, please ensure file exists")

    if not rsd.file_exists(rsd.conflation_file):
        raise FileNotFoundError(f"cannot find conflation file {rsd.conflation_file}, please ensure file exists")

    ripple1d_parameters = rsd.nwm_conflation_parameters(str(nwm_id))
    if ripple1d_parameters["us_xs"]["xs_id"] == "-9999":
        ripple1d_parameters["messages"] = f"skipping {nwm_id}; no cross sections conflated."
        logging.warning(ripple1d_parameters["messages"])

    else:
        rgs = RippleGeopackageSubsetter(rsd.ras_gpkg_file, rsd.conflation_file, submodel_directory, nwm_id)
        rgs.write_ripple_gpkg()
        ripple1d_parameters = rgs.update_ripple1d_parameters(rsd)
        rgs.write_ripple1d_parameters(ripple1d_parameters)

    return ripple1d_parameters
